from .operations import *

OPS = {
    'ZERO': lambda indim, outdim, dropout, concat=False: Zero(indim, outdim),
    'IDEN': lambda indim, outdim, dropout, concat=False: Identity(),
    'GCN': lambda indim, outdim, dropout, concat=False: GCNConv(indim, outdim, add_self_loops=False),
    'SAGE-MEAN': lambda indim, outdim, dropout, concat=False: SAGEConv(indim, outdim),
    'GAT16': lambda indim, outdim, dropout, concat=False: GATConv(indim, outdim, dropout=dropout, heads=16, concat=False, add_self_loops=False) if not concat else GATConv(indim, outdim // 16, dropout=dropout, heads=16, concat=True, add_self_loops=False),
    'GAT2': lambda indim, outdim, dropout, concat=False: GATConv(indim, outdim, dropout=dropout, heads=2, concat=False, add_self_loops=False) if not concat else GATConv(indim, outdim // 2, dropout=dropout, heads=2, concat=True, add_self_loops=False),
    'GAT4': lambda indim, outdim, dropout, concat=False: GATConv(indim, outdim, dropout=dropout, heads=4, concat=False, add_self_loops=False) if not concat else GATConv(indim, outdim // 4, dropout=dropout, heads=4, concat=True, add_self_loops=False),
    'GAT8': lambda indim, outdim, dropout, concat=False: GATConv(indim, outdim, dropout=dropout, heads=8, concat=False, add_self_loops=False) if not concat else GATConv(indim, outdim // 8, dropout=dropout, heads=8, concat=True, add_self_loops=False),
    'GAT1': lambda indim, outdim, dropout, concat=False: GATConv(indim, outdim, dropout=dropout, heads=1, concat=False, add_self_loops=False),
    'LIN': lambda indim, outdim, dropout, concat=False: Linear(indim, outdim),
    'ARMA': lambda indim, outdim, dropout, concat=False: ARMAConv(indim, outdim),
    'CHEB': lambda indim, outdim, dropout, concat=False: ChebConv(indim, outdim, 2),
    'SGC': lambda indim, outdim, dropout, concat=False: SGConv(indim, outdim, add_self_loops=False)
}

PRIMITIVES = list(OPS.keys())



import torch
import typing as _typing
import autogl
from ... import GeneralStaticGraph


class StaticGraphToGeneralData:
    def __init__(self, *__args, **__kwargs):
        pass

    def __call__(
            self, static_graph: GeneralStaticGraph,
            *__args, **__kwargs
    ):
        if not isinstance(static_graph, GeneralStaticGraph):
            raise TypeError
        elif not static_graph.nodes.is_homogeneous:
            raise ValueError("Provided static graph MUST consist of homogeneous nodes")
        homogeneous_node_type: _typing.Optional[str] = (
            list(static_graph.nodes)[0]
            if len(list(static_graph.nodes)) > 0 else None
        )
        data: _typing.Dict[str, torch.Tensor] = dict()
        if isinstance(homogeneous_node_type, str):
            node_and_edge_data_keys_intersection: _typing.Set[str] = (
                    set(static_graph.nodes.data) & set(static_graph.data)
            )
            if len(node_and_edge_data_keys_intersection) > 0:
                raise ValueError(
                    f"Provided static graph contains duplicate data "
                    f"with same keys {node_and_edge_data_keys_intersection}"
                    f"for homogeneous nodes data and graph-level data, "
                    f"please refer to doc for more details."
                )
            data.update(static_graph.nodes.data)
            data.update(static_graph.data)
        else:
            data.update(static_graph.data)

        if len(list(static_graph.edges)) == 1:
            data["edge_index"] = static_graph.edges.connections
            if len(set(data.keys()) & set(static_graph.edges.data.keys())) > 0:
                raise ValueError(
                    "Provided static graph contains duplicate data with same key, "
                    "please refer to doc for more details."
                )
            data.update(static_graph.edges.data)
        elif len(list(static_graph.edges)) > 1:
            for canonical_edge_type in static_graph.edges:
                if homogeneous_node_type is not None and isinstance(homogeneous_node_type, str) and (
                        canonical_edge_type.source_node_type != homogeneous_node_type or
                        canonical_edge_type.target_node_type != homogeneous_node_type
                ):
                    continue
                if len(canonical_edge_type.relation_type) < 4 or canonical_edge_type[-4:] != 'edge':
                    continue
                data[f"{canonical_edge_type.relation_type}_index"] = (
                    static_graph.edges[canonical_edge_type].connections
                )

                edge_type_prefix: str = canonical_edge_type.relation_type[:-4]
                for data_key in static_graph.edges[canonical_edge_type].data:
                    if len(data_key) >= 4 and data_key[:4] == 'edge':
                        data[f"{edge_type_prefix}{data_key}"] = (
                            static_graph.edges[canonical_edge_type].data[data_key].detach()
                        )
                    else:
                        data[f"{canonical_edge_type.relation_type}_{data_key}"] = (
                            static_graph.edges[canonical_edge_type].data[data_key].detach()
                        )

        general_data = autogl.data.Data()
        for key, value in data.items():
            setattr(general_data, key, value)
        return general_data


def static_graph_to_general_data(static_graph: GeneralStaticGraph) -> autogl.data.Data:
    return StaticGraphToGeneralData().__call__(static_graph)



import typing as _typing
from ._dataset import InMemoryDataset
from ..graph import GeneralStaticGraph


class InMemoryStaticGraphSet(InMemoryDataset[GeneralStaticGraph]):
    def __init__(
            self, graphs: _typing.Iterable[GeneralStaticGraph],
            train_index: _typing.Optional[_typing.Iterable[int]] = ...,
            val_index: _typing.Optional[_typing.Iterable[int]] = ...,
            test_index: _typing.Optional[_typing.Iterable[int]] = ...
    ):
        super(InMemoryStaticGraphSet, self).__init__(
            graphs, train_index, val_index, test_index
        )

    def __iter__(self) -> _typing.Iterator[GeneralStaticGraph]:
        return super(InMemoryStaticGraphSet, self).__iter__()

    def __getitem__(self, index: int) -> GeneralStaticGraph:
        return super(InMemoryStaticGraphSet, self).__getitem__(index)

    def __setitem__(self, index: int, data: GeneralStaticGraph):
        super(InMemoryStaticGraphSet, self).__setitem__(index, data)



"""
Auto Classfier for Graph Node Classification
"""
import time
import json

from copy import deepcopy
from typing import Sequence

import torch
import numpy as np
import yaml

from .utils import _parse_hp_space, _parse_model_hp, _initialize_single_model
from ..base import BaseClassifier
from ....module.model import EncoderUniversalRegistry, DecoderUniversalRegistry, ModelUniversalRegistry
from ....module.feature import FEATURE_DICT
from ....module.train import TRAINER_DICT, get_feval, BaseGraphClassificationTrainer
from ....module.train.ssl import BaseContrastiveTrainer
from ...utils import LeaderBoard, get_dataset_labels, set_seed, get_graph_from_dataset, get_graph_node_features, convert_dataset
from ....datasets import utils
from ...utils import get_logger
from ....backend import DependentBackend

LOGGER = get_logger("SSLGraphClassifier")
BACKEND = DependentBackend.get_backend_name()

class SSLGraphClassifier(BaseClassifier):
    """
    Auto Multi-class Graph Classifier.

    Used to automatically solve the graph classification problems.

    Parameters
    ----------
    feature_module: autogl.module.feature.BaseFeatureEngineer or str or None
        The (name of) auto feature engineer used to process the given dataset.
        Disable feature engineer by setting it to ``None``. Default ``deepgl``.

    graph_models: list of autogl.module.model.BaseModel or list of str
        The (name of) models to be optimized as backbone. Default ``['gat', 'gcn']``.

    hpo_module: autogl.module.hpo.BaseHPOptimizer or str or None
        The (name of) hpo module used to search for best hyper parameters.
        Disable hpo by setting it to ``None``. Default ``anneal``.

    ensemble_module: autogl.module.ensemble.BaseEnsembler or str or None
        The (name of) ensemble module used to ensemble the multi-models found.
        Disable ensemble by setting it to ``None``. Default ``voting``.

    max_evals: int (Optional)
        If given, will set the number eval times the hpo module will use.
        Only be effective when hpo_module is ``str``. Default ``None``.

    default_trainer: str (Optional)
        The (name of) the trainer used in this solver. Default to ``NodeClassificationFull``.

    trainer_hp_space: Iterable[dict] (Optional)
        trainer hp space or list of trainer hp spaces configuration.
        If a single trainer hp is given, will specify the hp space of trainer for
        every model. If a list of trainer hp is given, will specify every model
        with corrsponding trainer hp space. Default ``None``.

    model_hp_spaces: list of list of dict (Optional)
        model hp space configuration.
        If given, will specify every hp space of every passed model. Default ``None``.
        If the encoder(-decoder) is passed, the space should be a dict containing keys "encoder"
        and "decoder", specifying the detailed encoder decoder hp spaces.

    size: int (Optional)
        The max models ensemble module will use. Default ``None``.

    device: torch.device or str
        The device where model will be running on. If set to ``auto``, will use gpu
        when available. You can also specify the device by directly giving ``gpu`` or
        ``cuda:0``, etc. Default ``auto``.
    """

    # pylint: disable=W0102

    def __init__(
        self,
        feature_module=None,
        graph_models=("gin", "topkpool"),
        hpo_module="anneal",
        ensemble_module="voting",
        max_evals=50,
        default_trainer="GraphClassificationFull",
        trainer_hp_space=None,
        model_hp_spaces=None,
        size=4,
        device="auto",
    ):

        super().__init__(
            feature_module=feature_module,
            graph_models=graph_models,
            nas_algorithms=None,
            nas_spaces=None,
            nas_estimators=None,
            hpo_module=hpo_module,
            ensemble_module=ensemble_module,
            max_evals=max_evals,
            default_trainer=default_trainer,
            trainer_hp_space=trainer_hp_space,
            model_hp_spaces=model_hp_spaces,
            size=size,
            device=device,
        )

        self.dataset = None

    def _init_graph_module(
        self,
        graph_models,
        num_classes,
        num_features,
        feval,
        device,
        loss,
        num_graph_features,
    ) -> "SSLGraphClassifier":
        # load graph network module
        self.graph_model_list = []

        for i, model in enumerate(graph_models):
            # init the trainer
            if not isinstance(model, BaseContrastiveTrainer):
                trainer = (
                    self._default_trainer if not isinstance(self._default_trainer, (tuple, list))
                    else self._default_trainer[i]
                )
                if isinstance(trainer, str):
                    if trainer not in TRAINER_DICT:
                        raise KeyError(f"Does not support trainer {trainer}")
                    trainer = TRAINER_DICT[trainer]()
                if isinstance(model, (tuple, list)):
                    trainer.encoder = model[0]
                    trainer.decoder = model[1]
                else:
                    trainer.encoder = model
            else:
                trainer = model

            # set model hp space
            if self._model_hp_spaces is not None:
                if self._model_hp_spaces[i] is not None:
                    if isinstance(self._model_hp_spaces[i], dict):
                        encoder_hp_space = self._model_hp_spaces[i].get('encoder', None)
                        decoder_hp_space = self._model_hp_spaces[i].get('decoder', None)
                        prediction_head_hp_space = self._model_hp_spaces[i].get('prediction_head', None)
                    else:
                        encoder_hp_space = self._model_hp_spaces[i]
                        decoder_hp_space = None
                    if encoder_hp_space is not None:
                        trainer.encoder.hyper_parameter_space = encoder_hp_space
                    if decoder_hp_space is not None:
                        trainer.decoder.hyper_parameter_space = decoder_hp_space
                    if prediction_head_hp_space is not None:
                        trainer.prediction_head.hyper_parameter_space = prediction_head_hp_space
            
            # set trainer hp space
            if self._trainer_hp_space is not None:
                if isinstance(self._trainer_hp_space[0], list):
                    current_hp_for_trainer = self._trainer_hp_space[i]
                else:
                    current_hp_for_trainer = self._trainer_hp_space
                trainer.hyper_parameter_space = current_hp_for_trainer

            trainer.num_features = num_features
            trainer.num_classes = num_classes
            trainer.num_graph_features = num_graph_features
            trainer.loss = loss
            trainer.feval = feval
            trainer.to(device)
            self.graph_model_list.append(trainer)

        return self


    # pylint: disable=arguments-differ
    def fit(
        self,
        dataset,
        time_limit=-1,
        inplace=False,
        train_split=None,
        val_split=None,
        evaluation_method="infer",
        seed=None,
    ) -> "AutoGraphClassifier":
        """
        Fit current solver on given dataset.

        Parameters
        ----------
        dataset: autogl.data.dataset
            The multi-graph dataset needed to fit on.

        time_limit: int
            The time limit of the whole fit process (in seconds). If set below 0, will ignore
            time limit. Default ``-1``.

        inplace: bool
            Whether we process the given dataset in inplace manner. Default ``False``.
            Set it to True if you want to save memory by modifying the given dataset directly.

        train_split: float or int (Optional)
            The train ratio (in ``float``) or number (in ``int``) of dataset. If you want to use
            default train/val/test split in dataset, please set this to ``None``.
            Default ``None``.

        val_split: float or int (Optional)
            The validation ratio (in ``float``) or number (in ``int``) of dataset. If you want to
            use default train/val/test split in dataset, please set this to ``None``.
            Default ``None``.

        evaluation_method: (list of) str autogl.module.train.evaluation
            A (list of) evaluation method for current solver. If ``infer``, will automatically
            determine. Default ``infer``.

        seed: int (Optional)
            The random seed. If set to ``None``, will run everything at random.
            Default ``None``.

        Returns
        -------
        self: autogl.solver.AutoGraphClassifier
            A reference of current solver.
        """

        set_seed(seed)

        num_classes = get_dataset_labels(dataset).max().item() + 1

        if time_limit < 0:
            time_limit = 3600 * 24
        time_begin = time.time()

        # initialize leaderboard
        if evaluation_method == "infer":
            if hasattr(dataset, "metric"):
                evaluation_method = [dataset.metric]
            else:
                if num_classes == 2:
                    evaluation_method = ["auc"]
                else:
                    evaluation_method = ["acc"]
        assert isinstance(evaluation_method, list)
        evaluator_list = get_feval(evaluation_method)

        self.leaderboard = LeaderBoard(
            [e.get_eval_name() for e in evaluator_list],
            {e.get_eval_name(): e.is_higher_better() for e in evaluator_list},
        )

        # set up the dataset
        if train_split is None and val_split is None:
            assert hasattr(dataset, "train_split") and hasattr(dataset, "val_split"), (
                "The dataset has no default train/val split! "
                "Please manually pass train and val ratio."
            )
            LOGGER.info("Use the default train/val/test ratio in given dataset")
            # if hasattr(dataset.train_split, "n_splits"):
            #    cross_validation = True

        elif train_split is not None and val_split is not None:
            utils.graph_random_splits(dataset, train_split, val_split, seed=seed)
        else:
            LOGGER.error(
                "Please set both train_split and val_split explicitly. Detect %s is None.",
                "train_split" if train_split is None else "val_split",
            )
            raise ValueError(
                "In consistent setting of train/val split. Detect {} is None.".format(
                    "train_split" if train_split is None else "val_split"
                )
            )

        # feature engineering
        if self.feature_module is not None:
            self.feature_module.fit(dataset.train_split)
            dataset = self.feature_module.transform(dataset, inplace=inplace)

        self.dataset = dataset
        
        # check whether the dataset has features.
        # currently we only support graph classification with features.
        
        feat = get_graph_node_features(get_graph_from_dataset(dataset))
        assert feat is not None, (
            "Does not support fit on non node-feature dataset!"
            " Please add node features to dataset or specify feature engineers that generate"
            " node features."
        )
        num_features = feat.size(-1)

        # initialize graph networks
        self._init_graph_module(
            self.gml,
            num_features=num_features,
            num_classes=num_classes,
            feval=evaluator_list,
            device=self.runtime_device,
            loss="NT_Xent" if not hasattr(dataset, "loss") else dataset.loss,
            num_graph_features=(0
            if not hasattr(dataset[0], "gf")
            else dataset[0].gf.size(1)) if BACKEND == 'pyg' else 
            (0 if 'gf' not in dataset[0].data else dataset[0].data['gf'].size(1)),
        )

        # train the models and tune hpo
        result_valid = []
        names = []
        for idx, model in enumerate(self.graph_model_list):
            if time_limit < 0:
                time_for_each_model = None
            else:
                time_for_each_model = (time_limit - time.time() + time_begin) / (
                    len(self.graph_model_list) - idx
                )
            if self.hpo_module is None:
                model.initialize()
                model.train(convert_dataset(dataset), True)
                optimized = model
            else:
                optimized, _ = self.hpo_module.optimize(
                    trainer=model, dataset=convert_dataset(dataset), time_limit=time_for_each_model
                )
            # to save memory, all the trainer derived will be mapped to cpu
            optimized.to(torch.device("cpu"))
            name = str(optimized) + "_idx%d" % (idx)
            names.append(name)
            performance_on_valid, _ = optimized.get_valid_score(return_major=False)
            result_valid.append(
                optimized.get_valid_predict_proba().detach().cpu().numpy()
            )
            self.leaderboard.insert_model_performance(
                name,
                dict(
                    zip(
                        [e.get_eval_name() for e in evaluator_list],
                        performance_on_valid,
                    )
                ),
            )
            self.trained_models[name] = optimized

        # fit the ensemble model
        if self.ensemble_module is not None:
            performance = self.ensemble_module.fit(
                result_valid,
                get_dataset_labels(dataset)[dataset.val_index].cpu().numpy(),
                names,
                evaluator_list,
                n_classes=num_classes,
            )
            self.leaderboard.insert_model_performance(
                "ensemble",
                dict(zip([e.get_eval_name() for e in evaluator_list], performance)),
            )

        return self

    def fit_predict(
        self,
        dataset,
        time_limit=-1,
        inplace=False,
        train_split=None,
        val_split=None,
        evaluation_method="infer",
        seed=None,
        use_ensemble=True,
        use_best=True,
        name=None,
    ) -> np.ndarray:
        """
        Fit current solver on given dataset and return the predicted value.

        Parameters
        ----------
        dataset: torch_geometric.data.dataset.Dataset
            The dataset needed to fit on. This dataset must have only one graph.

        time_limit: int
            The time limit of the whole fit process (in seconds). If set below 0, will
            ignore time limit. Default ``-1``.

        inplace: bool
            Whether we process the given dataset in inplace manner. Default ``False``.
            Set it to True if you want to save memory by modifying the given dataset directly.

        train_split: float or int (Optional)
            The train ratio (in ``float``) or number (in ``int``) of dataset. If you want to
            use default train/val/test split in dataset, please set this to ``None``.
            Default ``None``.

        val_split: float or int (Optional)
            The validation ratio (in ``float``) or number (in ``int``) of dataset. If you want
            to use default train/val/test split in dataset, please set this to ``None``.
            Default ``None``.

        evaluation_method: (list of) str or autogl.module.train.evaluation
            A (list of) evaluation method for current solver. If ``infer``, will automatically
            determine. Default ``infer``.

        seed: int (Optional)
            The random seed. If set to ``None``, will run everything at random.
            Default ``None``.

        use_ensemble: bool
            Whether to use ensemble to do the predict. Default ``True``.

        use_best: bool
            Whether to use the best single model to do the predict. Will only be effective when
            ``use_ensemble`` is ``False``. Default ``True``.

        name: str or None
            The name of model used to predict. Will only be effective when ``use_ensemble`` and
            ``use_best`` both are ``False``. Default ``None``.

        Returns
        -------
        result: np.ndarray
            An array of shape ``(N,)``, where ``N`` is the number of test nodes. The prediction
            on given dataset.
        """
        self.fit(
            dataset=dataset,
            time_limit=time_limit,
            inplace=inplace,
            train_split=train_split,
            val_split=val_split,
            evaluation_method=evaluation_method,
            seed=seed,
        )
        return self.predict(
            dataset=dataset,
            inplaced=inplace,
            inplace=inplace,
            use_ensemble=use_ensemble,
            use_best=use_best,
            name=name,
        )

    def predict_proba(
        self,
        dataset=None,
        inplaced=False,
        inplace=False,
        use_ensemble=True,
        use_best=True,
        name=None,
        mask="test",
    ) -> np.ndarray:
        """
        Predict the node probability.

        Parameters
        ----------
        dataset: autogl.data.Dataset or None
            The dataset needed to predict. If ``None``, will use the processed dataset
            passed to ``fit()`` instead. Default ``None``.

        inplaced: bool
            Whether the given dataset is processed. Only be effective when ``dataset``
            is not ``None``. If you pass the dataset to ``fit()`` with ``inplace=True``,
            and you pass the dataset again to this method, you should set this argument
            to ``True``. Otherwise ``False``. Default ``False``.

        inplace: bool
            Whether we process the given dataset in inplace manner. Default ``False``.
            Set it to True if you want to save memory by modifying the given dataset directly.

        use_ensemble: bool
            Whether to use ensemble to do the predict. Default ``True``.

        use_best: bool
            Whether to use the best single model to do the predict. Will only be effective when
            ``use_ensemble`` is ``False``. Default ``True``.

        name: str or None
            The name of model used to predict. Will only be effective when ``use_ensemble`` and
            ``use_best`` both are ``False``. Default ``None``.

        mask: str
            The data split to give prediction on. Default ``test``.

        Returns
        -------
        result: np.ndarray
            An array of shape ``(N,C,)``, where ``N`` is the number of test nodes and ``C`` is
            the number of classes. The prediction on given dataset.
        """
        if dataset is None:
            dataset = self.dataset
        elif not inplaced:
            if self.feature_module is not None:
                dataset = self.feature_module.transform(dataset, inplace=inplace)

        if use_ensemble:
            LOGGER.info("Ensemble argument on, will try using ensemble model.")

        if not use_ensemble and use_best:
            LOGGER.info(
                "Ensemble argument off and best argument on, will try using best model."
            )

        if (use_ensemble and self.ensemble_module is not None) or (
            not use_best and name == "ensemble"
        ):
            # we need to get all the prediction of every model trained
            predict_result = []
            names = []
            for model_name in self.trained_models:
                predict_result.append(
                    self._predict_proba_by_name(dataset, model_name, mask)
                )
                names.append(model_name)
            return self.ensemble_module.ensemble(predict_result, names)

        if use_ensemble and self.ensemble_module is None:
            LOGGER.warning(
                "Cannot use ensemble because no ensebmle module is given. "
                "Will use best model instead."
            )

        if use_best or (use_ensemble and self.ensemble_module is None):
            # just return the best model we have found
            best_model_name = self.leaderboard.get_best_model()
            return self._predict_proba_by_name(dataset, best_model_name, mask)

        if name is not None:
            # return model performance by name
            return self._predict_proba_by_name(dataset, name, mask)

        LOGGER.error(
            "No model name is given while ensemble and best arguments are off."
        )
        raise ValueError(
            "You need to specify a model name if you do not want use ensemble and best model."
        )

    def _predict_proba_by_name(self, dataset, name, mask):
        self.trained_models[name].to(self.runtime_device)
        predicted = (
            self.trained_models[name]
            .predict_proba(convert_dataset(dataset), mask=mask)
            .detach()
            .cpu()
            .numpy()
        )
        self.trained_models[name].to(torch.device("cpu"))
        return predicted

    def predict(
        self,
        dataset=None,
        inplaced=False,
        inplace=False,
        use_ensemble=True,
        use_best=True,
        name=None,
        mask="test",
    ) -> np.ndarray:
        """
        Predict the node class number.

        Parameters
        ----------
        dataset: autogl.data.Dataset or None
            The dataset needed to predict. If ``None``, will use the processed dataset passed
            to ``fit()`` instead. Default ``None``.

        inplaced: bool
            Whether the given dataset is processed. Only be effective when ``dataset``
            is not ``None``. If you pass the dataset to ``fit()`` with ``inplace=True``, and
            you pass the dataset again to this method, you should set this argument to ``True``.
            Otherwise ``False``. Default ``False``.

        inplace: bool
            Whether we process the given dataset in inplace manner. Default ``False``.
            Set it to True if you want to save memory by modifying the given dataset directly.

        use_ensemble: bool
            Whether to use ensemble to do the predict. Default ``True``.

        use_best: bool
            Whether to use the best single model to do the predict. Will only be effective
            when ``use_ensemble`` is ``False``. Default ``True``.

        name: str or None
            The name of model used to predict. Will only be effective when ``use_ensemble``
            and ``use_best`` both are ``False``. Default ``None``.

        Returns
        -------
        result: np.ndarray
            An array of shape ``(N,)``, where ``N`` is the number of test nodes.
            The prediction on given dataset.
        """
        proba = self.predict_proba(
            dataset, inplaced, inplace, use_ensemble, use_best, name, mask
        )
        return np.argmax(proba, axis=1)

    def evaluate(self, dataset=None,
        inplaced=False,
        inplace=False,
        use_ensemble=True,
        use_best=True,
        name=None,
        mask="test",
        label=None,
        metric="acc"
    ):
        """
        Evaluate the given dataset.


        Parameters
        ----------
        dataset: torch_geometric.data.dataset.Dataset or None
            The dataset needed to predict. If ``None``, will use the processed dataset passed
            to ``fit()`` instead. Default ``None``.

        inplaced: bool
            Whether the given dataset is processed. Only be effective when ``dataset``
            is not ``None``. If you pass the dataset to ``fit()`` with ``inplace=True``, and
            you pass the dataset again to this method, you should set this argument to ``True``.
            Otherwise ``False``. Default ``False``.

        inplace: bool
            Whether we process the given dataset in inplace manner. Default ``False``. Set it to
            True if you want to save memory by modifying the given dataset directly.

        use_ensemble: bool
            Whether to use ensemble to do the predict. Default ``True``.

        use_best: bool
            Whether to use the best single model to do the predict. Will only be effective when
            ``use_ensemble`` is ``False``. Default ``True``.

        name: str or None
            The name of model used to predict. Will only be effective when ``use_ensemble`` and
            ``use_best`` both are ``False``. Default ``None``.

        mask: str
            The data split to give prediction on. Default ``test``.

        label: torch.Tensor (Optional)
            The groud truth label of the given predicted dataset split. If not passed, will extract
            labels from the input dataset.
        
        metric: str
            The metric to be used for evaluating the model. Default ``acc``.

        Returns
        -------
        score(s): (list of) evaluation scores
            the evaluation results according to the evaluator passed.

        """
        predicted = self.predict_proba(dataset, inplaced, inplace, use_ensemble, use_best, name, mask)
        if dataset is None:
            dataset = self.dataset
        if label is None:
            if mask == "all":
                masked_dataset = dataset
            else:
                masked_dataset = utils.graph_get_split(dataset, mask, False)
            label = get_dataset_labels(masked_dataset)
        evaluator = get_feval(metric)
        if isinstance(evaluator, Sequence):
            return [evals.evaluate(predicted, label) for evals in evaluator]
        return evaluator.evaluate(predicted, label)

    @classmethod
    def from_config(cls, path_or_dict, filetype="auto") -> "AutoGraphClassifier":
        """
        Load solver from config file.

        You can use this function to directly load a solver from predefined config dict
        or config file path. Currently, only support file type of ``json`` or ``yaml``,
        if you pass a path.

        Parameters
        ----------
        path_or_dict: str or dict
            The path to the config file or the config dictionary object

        filetype: str
            The filetype the given file if the path is specified. Currently only support
            ``json`` or ``yaml``. You can set to ``auto`` to automatically detect the file
            type (from file name). Default ``auto``.

        Returns
        -------
        solver: autogl.solver.AutoGraphClassifier
            The solver that is created from given file or dictionary.
        """
        assert filetype in ["auto", "yaml", "json"], (
            "currently only support yaml file or json file type, but get type "
            + filetype
        )
        if isinstance(path_or_dict, str):
            if filetype == "auto":
                if path_or_dict.endswith(".yaml") or path_or_dict.endswith(".yml"):
                    filetype = "yaml"
                elif path_or_dict.endswith(".json"):
                    filetype = "json"
                else:
                    LOGGER.error(
                        "cannot parse the type of the given file name, "
                        "please manually set the file type"
                    )
                    raise ValueError(
                        "cannot parse the type of the given file name, "
                        "please manually set the file type"
                    )
            if filetype == "yaml":
                path_or_dict = yaml.load(
                    open(path_or_dict, "r").read(), Loader=yaml.FullLoader
                )
            else:
                path_or_dict = json.load(open(path_or_dict, "r"))

        # load the dictionary
        path_or_dict = deepcopy(path_or_dict)
        solver = cls(None, [], None, None)
        fe_list = path_or_dict.pop("feature", None)
        if fe_list is not None:
            fe_list_ele = []
            for feature_engineer in fe_list:
                name = feature_engineer.pop("name")
                if name is not None:
                    fe_list_ele.append(FEATURE_DICT[name](**feature_engineer))
            if fe_list_ele != []:
                solver.set_feature_module(fe_list_ele)

        models = path_or_dict.pop("models", [{"name": "gcn"}, {"name": "gat"}, {"name": "sage"}, {"name": "gin"}])
        # models should be a list of model
        # with each element in two cases
        # * a dict describing a certain model
        # * a dict containing {"encoder": encoder, "decoder": decoder}
        model_hp_space = [
            _parse_model_hp(model) for model in models
        ]
        model_list = [
            _initialize_single_model(model) for model in models
        ]

        trainer = path_or_dict.pop("trainer", None)
        default_trainer = "GraphClassificationFull"
        trainer_space = None
        if isinstance(trainer, dict):
            # global default
            default_trainer = trainer.pop("name", "GraphClassificationFull")
            trainer_space = _parse_hp_space(trainer.pop("hp_space", None))
            default_kwargs = {"num_features": None, "num_classes": None}
            default_kwargs.update(trainer)
            default_kwargs["init"] = False
            for i in range(len(model_list)):
                model = model_list[i]
                trainer_wrapper = TRAINER_DICT[default_trainer](
                    model=model, prediction_head=models[i]["prediction_head"].pop("name", None), **default_kwargs
                )
                model_list[i] = trainer_wrapper
        elif isinstance(trainer, list):
            # sequential trainer definition
            assert len(trainer) == len(
                model_list
            ), "The number of trainer and model does not match"
            trainer_space = []
            for i in range(len(model_list)):
                train, model = trainer[i], model_list[i]
                default_trainer = train.pop("name", "GraphClassificationFull")
                trainer_space.append(_parse_hp_space(train.pop("hp_space", None)))
                default_kwargs = {"num_features": None, "num_classes": None}
                default_kwargs.update(train)
                default_kwargs["init"] = False
                trainer_wrap = TRAINER_DICT[default_trainer](
                    model=model, **default_kwargs
                )
                model_list[i] = trainer_wrap

        solver.set_graph_models(
            model_list, default_trainer, trainer_space, model_hp_space
        )

        hpo_dict = path_or_dict.pop("hpo", {"name": "anneal"})
        if hpo_dict is not None:
            name = hpo_dict.pop("name")
            solver.set_hpo_module(name, **hpo_dict)

        ensemble_dict = path_or_dict.pop("ensemble", {"name": "voting"})
        if ensemble_dict is not None:
            name = ensemble_dict.pop("name")
            solver.set_ensemble_module(name, **ensemble_dict)

        return solver



import dgl
import torch
import typing as _typing
from autogl.data import Dataset, InMemoryDataset
from autogl.data.graph import GeneralStaticGraph
from autogl.data.graph.utils import conversion


def to_dgl_dataset(
        dataset: _typing.Union[Dataset, _typing.Iterable[GeneralStaticGraph]]
) -> Dataset[_typing.Union[dgl.DGLGraph, _typing.Tuple[dgl.DGLGraph, torch.Tensor]]]:
    def _transform(
            general_static_graph: GeneralStaticGraph
    ) -> _typing.Union[dgl.DGLGraph, _typing.Tuple[dgl.DGLGraph, torch.Tensor]]:
        if not isinstance(general_static_graph, GeneralStaticGraph):
            raise TypeError
        if 'label' in general_static_graph.data:
            label: _typing.Optional[torch.Tensor] = general_static_graph.data['label']
        elif 'y' in general_static_graph.data:
            label: _typing.Optional[torch.Tensor] = general_static_graph.data['y']
        else:
            label: _typing.Optional[torch.Tensor] = None
        if label is not None and isinstance(label, torch.Tensor) and torch.is_tensor(label):
            return conversion.general_static_graph_to_dgl_graph(general_static_graph), label
        else:
            return conversion.general_static_graph_to_dgl_graph(general_static_graph)

    transformed_datalist: _typing.MutableSequence[
        _typing.Union[dgl.DGLGraph, _typing.Tuple[dgl.DGLGraph, torch.Tensor]]
    ] = []
    for item in dataset:
        if isinstance(item, GeneralStaticGraph):
            transformed_datalist.append(_transform(item))
        elif isinstance(item, dgl.DGLGraph):
            transformed_datalist.append(item)
        elif (
                isinstance(item, _typing.Sequence) and len(item) == 2 and
                isinstance(item[0], dgl.DGLGraph) and isinstance(item[1], torch.Tensor)
        ):
            transformed_datalist.append(tuple(item))
        else:
            raise ValueError(f"Illegal data item as {item}")

    return (
        InMemoryDataset(transformed_datalist, dataset.train_index, dataset.val_index, dataset.test_index, dataset.schema)
        if isinstance(dataset, InMemoryDataset)
        else InMemoryDataset(transformed_datalist)
    )



from .base_encoder import BaseEncoderMaintainer, AutoHomogeneousEncoderMaintainer
from .encoder_registry import EncoderUniversalRegistry
from autogl.backend import DependentBackend

if DependentBackend.is_pyg():
    from ._pyg import (
        GCNEncoderMaintainer,
        GATEncoderMaintainer,
        GINEncoderMaintainer,
        SAGEEncoderMaintainer
    )
else:
    from ._dgl import (
        GCNEncoderMaintainer,
        GATEncoderMaintainer,
        GINEncoderMaintainer,
        SAGEEncoderMaintainer,
        AutoTopKMaintainer
    )

__all__ = [
    "BaseEncoderMaintainer",
    "EncoderUniversalRegistry",
    "AutoHomogeneousEncoderMaintainer",
    "GCNEncoderMaintainer",
    "GATEncoderMaintainer",
    "GINEncoderMaintainer",
    "SAGEEncoderMaintainer",
]

if DependentBackend.is_dgl():
    __all__.append("AutoTopKMaintainer")



import dgl
import torch
import typing as _typing
from ..._general_static_graph import GeneralStaticGraph
from ... import (
    _general_static_graph_generator, _general_static_graph_dgl_implementation
)


class GeneralStaticGraphToDGLGraph:
    def __init__(self, *__args, **__kwargs):
        pass

    def __call__(self, static_graph: GeneralStaticGraph) -> dgl.DGLGraph:
        dgl_graph: dgl.DGLGraph = dgl.heterograph(
            dict([
                (
                    (
                        canonical_edge_type.source_node_type,
                        canonical_edge_type.relation_type,
                        canonical_edge_type.target_node_type
                    ),
                    (
                        static_graph.edges[canonical_edge_type].connections[0],
                        static_graph.edges[canonical_edge_type].connections[1]
                    )
                )
                for canonical_edge_type in static_graph.edges
            ])
        )
        for node_type in static_graph.nodes:
            for data_key in static_graph.nodes[node_type].data:
                dgl_graph.nodes[node_type].data[data_key] = (
                    static_graph.nodes[node_type].data[data_key]
                )
        for canonical_edge_type in static_graph.edges:
            for data_key in static_graph.edges[canonical_edge_type].data:
                dgl_graph.edges[
                    (
                        canonical_edge_type.source_node_type,
                        canonical_edge_type.relation_type,
                        canonical_edge_type.target_node_type
                    )
                ].data[data_key] = (
                    static_graph.edges[canonical_edge_type].data[data_key]
                )
        # Set graph level data by `setattr`
        if len(static_graph.data) > 0:
            setattr(dgl_graph, "graph_data", dict(static_graph.data))
            if "gf" in static_graph.data:
                setattr(dgl_graph, "gf", static_graph.data["gf"].detach().clone())
        return dgl_graph


class DGLGraphToGeneralStaticGraph:
    def __init__(
            self, as_universal_storage_format: bool = False,
            *__args, **__kwargs
    ):
        if not isinstance(as_universal_storage_format, bool):
            raise TypeError
        else:
            self._as_universal_storage_format: bool = as_universal_storage_format

    def __call__(
            self, dgl_graph: dgl.DGLGraph,
            as_universal_storage_format: _typing.Optional[bool] = ...,
            *__args, **__kwargs
    ) -> GeneralStaticGraph:
        if not (
                as_universal_storage_format in (Ellipsis, None) or
                isinstance(as_universal_storage_format, bool)
        ):
            raise TypeError
        _as_universal_storage_format: bool = (
            as_universal_storage_format
            if isinstance(as_universal_storage_format, bool)
            else self._as_universal_storage_format
        )

        if not _as_universal_storage_format:
            general_static_graph: GeneralStaticGraph = (
                _general_static_graph_dgl_implementation.GeneralStaticGraphDGLImplementation(dgl_graph)
            )

        else:
            general_static_graph: GeneralStaticGraph = (
                _general_static_graph_generator.GeneralStaticGraphGenerator.create_heterogeneous_static_graph(
                    dict([(node_type, dgl_graph.nodes[node_type].data) for node_type in dgl_graph.ntypes]),
                    dict([
                        (
                            canonical_edge_type,
                            (
                                torch.vstack(dgl_graph.edges(etype=canonical_edge_type)),
                                dgl_graph.edges[canonical_edge_type].data
                            )
                        )
                        for canonical_edge_type in dgl_graph.canonical_etypes]
                    )
                )
            )
        if (
                hasattr(dgl_graph, "graph_data") and
                isinstance(getattr(dgl_graph, "graph_data"), _typing.Mapping)
        ):
            graph_data: _typing.Mapping[str, torch.Tensor] = getattr(dgl_graph, "graph_data")
            for k, v in graph_data.items():
                if (
                        isinstance(k, str) and ' ' not in k and
                        isinstance(v, torch.Tensor)
                ):
                    general_static_graph.data[k] = v
        for k in ("gf",):
            if (
                    hasattr(dgl_graph, k) and
                    isinstance(getattr(dgl_graph, k), torch.Tensor)
            ):
                general_static_graph.data[k] = getattr(dgl_graph, k)
        return general_static_graph


def general_static_graph_to_dgl_graph(
        general_static_graph: GeneralStaticGraph, *__args, **__kwargs
) -> dgl.DGLGraph:
    return GeneralStaticGraphToDGLGraph(*__args, **__kwargs).__call__(
        general_static_graph
    )


def dgl_graph_to_general_static_graph(
        dgl_graph: dgl.DGLGraph, as_universal_storage_format: bool = False,
        *__args, **__kwargs
) -> GeneralStaticGraph:
    return DGLGraphToGeneralStaticGraph(as_universal_storage_format).__call__(
        dgl_graph, as_universal_storage_format
    )



from . import register_nas_estimator
from ..space import BaseSpace
from .base import BaseEstimator
from .one_shot import OneShotEstimator, OneShotEstimator_HardwareAware
import torch

from autogl.module.train import NodeClassificationFullTrainer, Acc


@register_nas_estimator("scratch")
class TrainEstimator(BaseEstimator):
    """
    An estimator which trans from scratch

    Parameters
    ----------
    loss_f : str
        The name of loss funciton in PyTorch
    evaluation : list of Evaluation
        The evaluation metrics in module/train/evaluation
    """

    def __init__(self, loss_f="nll_loss", evaluation=[Acc()]):
        super().__init__(loss_f, evaluation)
        self.evaluation = evaluation
        self.estimator = OneShotEstimator(self.loss_f, self.evaluation)

    def infer(self, model: BaseSpace, dataset, mask="train"):
        boxmodel = model.wrap()
        self.trainer = NodeClassificationFullTrainer(
            model=boxmodel,
            optimizer=torch.optim.Adam,
            lr=0.005,
            max_epoch=300,
            early_stopping_round=30,
            weight_decay=5e-4,
            device="auto",
            init=False,
            feval=self.evaluation,
            loss=self.loss_f,
            lr_scheduler_type=None,
        )
        try:
            self.trainer.train(dataset)
            with torch.no_grad():
                return self.estimator.infer(boxmodel.model, dataset, mask)
        except RuntimeError as e:
            if "cuda" in str(e) or "CUDA" in str(e):
                INF = 100
                fin = [-INF if eva.is_higher_better else INF for eva in self.evaluation]
                return fin, 0
            else:
                raise e


@register_nas_estimator("scratch_hardware")
class TrainEstimator_HardwareAware(TrainEstimator):
    """
    An hardware-aware estimator which trans from scratch
    """

    def __init__(
        self,
        loss_f="nll_loss",
        evaluation=[Acc()],
        hardware_evaluation="parameter",
        hardware_metric_weight=0,
    ):
        super().__init__(loss_f, evaluation)
        self.estimator = OneShotEstimator_HardwareAware(
            self.loss_f, self.evaluation, hardware_evaluation, hardware_metric_weight
        )



import numpy as np
import scipy.sparse as sp
import torch

# for GNNGuard
def accuracy(output, labels):
    """Return accuracy of output compared to labels.

    Parameters
    ----------
    output : torch.Tensor
        output from model
    labels : torch.Tensor or numpy.array
        node labels

    Returns
    -------
    float
        accuracy
    """
    if not hasattr(labels, '__len__'):
        labels = [labels]
    if type(labels) is not torch.Tensor:
        labels = torch.LongTensor(labels)
    preds = output.max(1)[1].type_as(labels)
    correct = preds.eq(labels).double()
    correct = correct.sum()
    return correct / len(labels)

# for GCNSVD
def normalize_adj(mx):
    """Normalize sparse adjacency matrix,
    A' = (D + I)^-1/2 * ( A + I ) * (D + I)^-1/2
    Row-normalize sparse matrix
    Parameters
    ----------
    mx : scipy.sparse.csr_matrix
        matrix to be normalized
    Returns
    -------
    scipy.sprase.lil_matrix
        normalized matrix
    """

    # TODO: maybe using coo format would be better?
    if type(mx) is not sp.lil.lil_matrix:
        mx = mx.tolil()
    if mx[0, 0] == 0 :
        mx = mx + sp.eye(mx.shape[0])
    rowsum = np.array(mx.sum(1))
    r_inv = np.power(rowsum, -1/2).flatten()
    r_inv[np.isinf(r_inv)] = 0.
    r_mat_inv = sp.diags(r_inv)
    mx = r_mat_inv.dot(mx)
    mx = mx.dot(r_mat_inv)
    return mx


def to_tensor(adj, features, labels=None, device='cpu'):
    """Convert adj, features, labels from array or sparse matrix to
    torch Tensor.
    Parameters
    ----------
    adj : scipy.sparse.csr_matrix
        the adjacency matrix.
    features : scipy.sparse.csr_matrix
        node features
    labels : numpy.array
        node labels
    device : str
        'cpu' or 'cuda'
    """
    if sp.issparse(adj):
        adj = sparse_mx_to_torch_sparse_tensor(adj)
    else:
        adj = torch.FloatTensor(adj)
    if sp.issparse(features):
        features = sparse_mx_to_torch_sparse_tensor(features)
    else:
        features = torch.FloatTensor(np.array(features))

    if labels is None:
        return adj.to(device), features.to(device)
    else:
        labels = torch.LongTensor(labels)
        return adj.to(device), features.to(device), labels.to(device)

def normalize_adj_tensor(adj, sparse=False):
    """Normalize adjacency tensor matrix.
    """
    device = torch.device("cuda" if adj.is_cuda else "cpu")
    if sparse:
        # warnings.warn('If you find the training process is too slow, you can uncomment line 207 in deeprobust/graph/utils.py. Note that you need to install torch_sparse')
        # TODO if this is too slow, uncomment the following code,
        # but you need to install torch_scatter
        # return normalize_sparse_tensor(adj)
        adj = to_scipy(adj)
        mx = normalize_adj(adj)
        return sparse_mx_to_torch_sparse_tensor(mx).to(device)
    else:
        mx = adj + torch.eye(adj.shape[0]).to(device)
        rowsum = mx.sum(1)
        r_inv = rowsum.pow(-1/2).flatten()
        r_inv[torch.isinf(r_inv)] = 0.
        r_mat_inv = torch.diag(r_inv)
        mx = r_mat_inv @ mx
        mx = mx @ r_mat_inv
    return mx

def to_scipy(tensor):
    """Convert a dense/sparse tensor to scipy matrix"""
    if is_sparse_tensor(tensor):
        values = tensor._values()
        indices = tensor._indices()
        return sp.csr_matrix((values.cpu().numpy(), indices.cpu().numpy()), shape=tensor.shape)
    else:
        indices = tensor.nonzero().t()
        values = tensor[indices[0], indices[1]]
        return sp.csr_matrix((values.cpu().numpy(), indices.cpu().numpy()), shape=tensor.shape)
def is_sparse_tensor(tensor):
    """Check if a tensor is sparse tensor.
    Parameters
    ----------
    tensor : torch.Tensor
        given tensor
    Returns
    -------
    bool
        whether a tensor is sparse tensor
    """
    # if hasattr(tensor, 'nnz'):
    if tensor.layout == torch.sparse_coo:
        return True
    else:
        return False

def sparse_mx_to_torch_sparse_tensor(sparse_mx):
    """Convert a scipy sparse matrix to a torch sparse tensor."""
    sparse_mx = sparse_mx.tocoo().astype(np.float32)
    sparserow=torch.LongTensor(sparse_mx.row).unsqueeze(1)
    sparsecol=torch.LongTensor(sparse_mx.col).unsqueeze(1)
    sparseconcat=torch.cat((sparserow, sparsecol),1)
    sparsedata=torch.FloatTensor(sparse_mx.data)
    return torch.sparse.FloatTensor(sparseconcat.t(),sparsedata,torch.Size(sparse_mx.shape))

	# slower version....
    # sparse_mx = sparse_mx.tocoo().astype(np.float32)
    # indices = torch.from_numpy(
    #     np.vstack((sparse_mx.row, sparse_mx.col)).astype(np.int64))
    # values = torch.from_numpy(sparse_mx.data)
    # shape = torch.Size(sparse_mx.shape)
    # return torch.sparse.FloatTensor(indices, values, shape)


import torch
from torch import nn
import torch.nn.functional as F
def act_map(act):
    if act == "linear":
        return lambda x: x
    elif act == "elu":
        return F.elu
    elif act == "sigmoid":
        return torch.sigmoid
    elif act == "tanh":
        return torch.tanh
    elif act == "relu":
        return torch.nn.functional.relu
    elif act == "relu6":
        return torch.nn.functional.relu6
    elif act == "softplus":
        return torch.nn.functional.softplus
    elif act == "leaky_relu":
        return torch.nn.functional.leaky_relu
    else:
        raise Exception("wrong activate function")

from ..backend import *
if is_dgl():
    from .operation_dgl import gnn_map,GeoLayer
else:
    from .operation_pyg import gnn_map,GeoLayer


"""
Ensemble module.
"""

import numpy as np

import torch
import torch.nn.functional as F
from sklearn.ensemble import GradientBoostingClassifier
from sklearn.linear_model import LogisticRegression

from .base import BaseEnsembler
from . import register_ensembler
from ...utils import get_logger

STACKING_LOGGER = get_logger("stacking")


@register_ensembler("stacking")
class Stacking(BaseEnsembler):
    """
    A stacking ensembler. Currently we support gradient boosting as the meta-algorithm.

    Parameters
    ----------
    meta_model : 'gbm' or 'glm' (Optional)
        Type of the stacker:
            'gbm' : Gradient boosting model. This is the default.
            'glm' : Generalized linear model.

    meta_params : a ``dict``  (Optional)
        When ``meta_model`` is specified, you can customize the parameters of the stacker.
        If this argument is not provided, the stacker will be configurated with default parameters.
        Default ``{}``.
    """

    def __init__(self, meta_model="gbm", meta_params={}, *args, **kwargs):
        super().__init__()
        self.model_name = meta_model.lower()
        assert self.model_name in [
            "gbm",
            "glm",
        ], "Only support gbm and glm when ensemble!"
        self.meta_params = meta_params

    def fit(
        self, predictions, label, identifiers, feval, n_classes="auto", *args, **kwargs
    ):
        """
        Fit the ensembler to the given data using Stacking method.

        Parameters
        ----------
        predictions : a list of np.ndarray
            Predictions of base learners (corresponding to the elements in identifiers).

        label : a list of int
            Class labels of instances.

        identifiers : a list of str
            The names of base models.

        feval : (a list of) autogl.module.train.evaluate
            Performance evaluation metrices.

        n_classes : int or str (Optional)
            The number of classes. Default as ``'auto'``, which will use maximum label.

        Returns
        -------
        (a list of) float
            The validation performance of the final stacker.
        """

        n_classes = n_classes if not n_classes == "auto" else max(label) + 1
        assert n_classes > max(
            label
        ), "Detect max label passed (%d) exceeeds" " n_classes given (%d)" % (
            max(label),
            n_classes,
        )

        assert len(identifiers) == len(
            set(identifiers)
        ), "Duplicate name" " in identifiers {} !".format(identifiers)

        self.fit_identifiers = identifiers

        if not isinstance(feval, list):
            feval = [feval]

        self._re_initialize(identifiers, len(predictions))

        config = self.meta_params

        STACKING_LOGGER.debug("meta-model name %s", self.model_name)

        if self.model_name == "gbm":
            meta_X = (
                torch.tensor(predictions).transpose(0, 1).flatten(start_dim=1).numpy()
            )
            meta_Y = np.array(label)
            config = {}
            model = GradientBoostingClassifier(**config)
            model.fit(meta_X, meta_Y)

            self.model = model
            ensemble_prediction = model.predict_proba(meta_X)

        elif self.model_name == "glm":
            meta_X = (
                torch.tensor(predictions).transpose(0, 1).flatten(start_dim=1).numpy()
            )
            meta_Y = np.array(label)

            config["multi_class"] = "auto"
            config["solver"] = "lbfgs"
            model = LogisticRegression(**config)
            model.fit(meta_X, meta_Y)

            self.model = model
            ensemble_prediction = model.predict_proba(meta_X)

        elif self.model_name == "nn":
            meta_X = torch.tensor(predictions).transpose(0, 1).flatten(start_dim=1)
            meta_Y = F.one_hot(
                torch.tensor(label, dtype=torch.int64), n_classes
            ).double()
            # print(meta_Y.type())

            n_instance, n_input = meta_X.size()
            n_learners = len(identifiers)

            fc = torch.nn.Linear(n_input, n_input // n_learners).double()

            config["lr"] = 1e-1
            # config['weight_decay'] = 1e-2
            optimizer = torch.optim.SGD(fc.parameters(), **config)

            max_epoch = 100
            for epoch in range(max_epoch):
                optimizer.zero_grad()
                ensemble_prediction = F.normalize(fc.forward(meta_X), dim=0)
                loss = F.mse_loss(ensemble_prediction, meta_Y)
                loss.backward()
                optimizer.step()

            self.model = fc
            ensemble_prediction = (
                F.normalize(fc.forward(meta_X), dim=0).detach().numpy()
            )

        else:
            STACKING_LOGGER.error(
                "Cannot parse stacking ensemble model name %s", self.model_name
            )

        return [fx.evaluate(ensemble_prediction, label) for fx in feval]

    def ensemble(self, predictions, identifiers, *args, **kwargs):
        """
        Ensemble the predictions of base models.

        Parameters
        ----------
        predictions : a list of ``np.ndarray``
            Predictions of base learners (corresponding to the elements in identifiers).
        identifiers : a list of ``str``
            The names of base models.

        Returns
        -------
        ``np.ndarray``
            The ensembled predictions.
        """

        assert len(identifiers) == len(
            set(identifiers)
        ), "Duplicate name in" " identifiers {} !".format(identifiers)

        assert set(self.fit_identifiers) == set(
            identifiers
        ), "Different identifiers" " passed in fit {} and ensemble {} !".format(
            self.fit_identifiers, identifiers
        )

        # re-order predictions if needed
        if not self.fit_identifiers == identifiers:
            re_id = [
                identifiers.index(identifier) for identifier in self.fit_identifiers
            ]
            predictions = [predictions[i] for i in re_id]

        if self.model_name in ["gbm", "glm"]:
            pred_packed = (
                torch.tensor(predictions).transpose(0, 1).flatten(start_dim=1).numpy()
            )
            return self.model.predict_proba(pred_packed)

        elif self.model_name in ["nn"]:
            pred_packed = torch.tensor(predictions).transpose(0, 1).flatten(start_dim=1)
            return F.normalize(self.model.forward(pred_packed), dim=0).detach().numpy()

    def _re_initialize(self, identifiers, n_models):
        self.identifiers = identifiers
        self.model = None



from abc import abstractmethod
import torch.nn as nn
from nni.nas.pytorch import mutables
from nni.nas.pytorch.fixed import FixedArchitecture
import json
from copy import deepcopy
import torch
from ...model import BaseAutoModel
from ....utils import get_logger
from ..utils import get_hardware_aware_metric


class OrderedMutable:
    """
    An abstract class with order, enabling to sort mutables with a certain rank.

    Parameters
    ----------
    order : int
        The order of the mutable
    """

    def __init__(self, order):
        self.order = order


class OrderedLayerChoice(OrderedMutable, mutables.LayerChoice):
    def __init__(
        self, order, op_candidates, reduction="sum", return_mask=False, key=None
    ):
        OrderedMutable.__init__(self, order)
        mutables.LayerChoice.__init__(
            self, op_candidates, reduction, return_mask, key)


class OrderedInputChoice(OrderedMutable, mutables.InputChoice):
    def __init__(
        self,
        order,
        n_candidates=None,
        choose_from=None,
        n_chosen=None,
        reduction="sum",
        return_mask=False,
        key=None,
    ):
        OrderedMutable.__init__(self, order)
        mutables.InputChoice.__init__(
            self, n_candidates, choose_from, n_chosen, reduction, return_mask, key
        )


class StrModule(nn.Module):
    """
    A shell used to wrap choices as nn.Module for non-one-shot space definition
    You can use ``map_nn`` function

    Parameters
    ----------
    name : anything
        the name of module, can be any type
    """

    def __init__(self, name):
        super().__init__()
        self.str = name

    def forward(self, *args, **kwargs):
        return self.str

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, self.str)


def map_nn(names):
    """
    A function used to wrap choices as nn.Module for non-one-shot space definition

    Parameters
    ----------
    name : list of anything
        the names of module, can be any type
    """
    return [StrModule(x) for x in names]


class BoxModel(BaseAutoModel):
    """
    The box wrapping a space, can be passed to later procedure or trainer

    Parameters
    ----------
    space_model : BaseSpace
        The space which should be wrapped
    device : str or torch.device
        The device to place the model
    """

    _logger = get_logger("space model")

    def __init__(self, space_model, device):
        super().__init__(None, None, device)
        self.init = True
        self.space = []
        self.hyperparams = {}
        self._model = space_model
        self.num_features = self._model.input_dim
        self.num_classes = self._model.output_dim
        self.params = {"num_class": self.num_classes, "features_num": self.num_features}
        self.selection = None

    def _initialize(self):
        return True

    def fix(self, selection):
        """
        To fix self._model with a selection

        Parameters
        ----------
        selection : dict
            A seletion indicating the choices of mutables
        """
        self.selection = selection
        self._model.instantiate()
        apply_fixed_architecture(self._model, selection, verbose=False)
        return self

    def forward(self, *args, **kwargs):
        return self._model(*args, **kwargs)

    def from_hyper_parameter(self, hp):
        """
        receive no hp, just copy self and reset the learnable parameters.
        """

        ret_self = deepcopy(self)
        ret_self._model.instantiate()
        if ret_self.selection:
            apply_fixed_architecture(ret_self._model, ret_self.selection, verbose=False)
        return ret_self

    def __repr__(self) -> str:
        return str(
            {'parameter': get_hardware_aware_metric(self.model, 'parameter'),
             'model': self.model,
             'selection': self.selection
             })

class BaseSpace(nn.Module):
    """
    Base space class of NAS module. Defining space containing all models.
    Please use mutables to define your whole space. Refer to
    `https://nni.readthedocs.io/en/stable/NAS/WriteSearchSpace.html`
    for detailed information.

    Parameters
    ----------
    init: bool
        Whether to initialize the whole space. Default: `False`
    """

    def __init__(self):
        super().__init__()
        self._initialized = False

    @abstractmethod
    def _instantiate(self):
        """
        Instantiate modules in the space
        """
        raise NotImplementedError()

    @abstractmethod
    def forward(self, *args, **kwargs):
        """
        Define the forward pass of space model
        """
        raise NotImplementedError()

    @abstractmethod
    def parse_model(self, selection: dict, device) -> BaseAutoModel:
        """
        Export the searched model from space.

        Parameters
        ----------
        selection: Dict
            The dictionary containing all the choices of nni.
        device: str or torch.device
            The device to put model on.

        Return
        ------
        model: autogl.module.model.BaseModel
            model to be exported.
        """
        raise NotImplementedError()

    def instantiate(self):
        """
        Instantiate the space, reset default key for the mutables here/
        """
        self._default_key = 0
        if not self._initialized:
            self._initialized = True

    def setLayerChoice(
        self, order, op_candidates, reduction="sum", return_mask=False, key=None
    ):
        """
        Give a unique key if not given
        """
        orikey = key
        if orikey == None:
            key = f"default_key_{self._default_key}"
            self._default_key += 1
            orikey = key
        layer = OrderedLayerChoice(
            order, op_candidates, reduction, return_mask, orikey)
        return layer

    def setInputChoice(
        self,
        order,
        n_candidates=None,
        choose_from=None,
        n_chosen=None,
        reduction="sum",
        return_mask=False,
        key=None,
    ):
        """
        Give a unique key if not given
        """
        orikey = key
        if orikey == None:
            key = f"default_key_{self._default_key}"
            self._default_key += 1
            orikey = key
        layer = OrderedInputChoice(
            order, n_candidates, choose_from, n_chosen, reduction, return_mask, orikey
        )
        return layer

    def wrap(self):
        """
        Return a BoxModel which wrap self as a model
        Used to pass to trainer
        To use this function, must contain `input_dim` and `output_dim`
        """
        device = next(self.parameters()).device
        return BoxModel(self, device)


class FixedInputChoice(nn.Module):
    """
    Use to replace `InputChoice` Mutable in fix process

    Parameters
    ----------
    mask : list
        The mask indicating which input to choose
    """

    def __init__(self, mask):
        self.mask_len = len(mask)
        for i in range(self.mask_len):
            if mask[i]:
                self.selected = i
                break
        super().__init__()

    def forward(self, optional_inputs):
        if len(optional_inputs) == self.mask_len:
            return optional_inputs[self.selected]


class CleanFixedArchitecture(FixedArchitecture):
    """
    Fixed architecture mutator that always selects a certain graph, allowing deepcopy

    Parameters
    ----------
    model : nn.Module
        A mutable network.
    fixed_arc : dict
        Preloaded architecture object.
    strict : bool
        Force everything that appears in ``fixed_arc`` to be used at least once.
    verbose : bool
        Print log messages if set to True
    """

    def __init__(self, model, fixed_arc, strict=True, verbose=True):
        super().__init__(model, fixed_arc, strict, verbose)

    def replace_all_choice(self, module=None, prefix=""):
        """
        Replace all choices with selected candidates. It's done with best effort.
        In case of weighted choices or multiple choices. if some of the choices on weighted with zero, delete them.
        If single choice, replace the module with a normal module.

        Parameters
        ----------
        module : nn.Module
            Module to be processed.
        prefix : str
            Module name under global namespace.
        """
        
        if module is None:
            module = self.model
        for name, mutable in module.named_children():
            global_name = (prefix + "." if prefix else "") + name
            if isinstance(mutable, OrderedLayerChoice):
                chosen = self._fixed_arc[mutable.key]
                if sum(chosen) == 1 and max(chosen) == 1 and not mutable.return_mask:
                    # sum is one, max is one, there has to be an only one
                    # this is compatible with both integer arrays, boolean arrays and float arrays
                    setattr(module, name, mutable[chosen.index(1)])
                else:
                    # remove unused parameters
                    for ch, n in zip(chosen, mutable.names):
                        if ch == 0 and not isinstance(ch, float):
                            setattr(mutable, n, None)
            elif isinstance(mutable, OrderedInputChoice):
                chosen = self._fixed_arc[mutable.key]
                setattr(module, name, FixedInputChoice(chosen))
            else:
                self.replace_all_choice(mutable, global_name)


def apply_fixed_architecture(model, fixed_arc, verbose=True):
    """
    Load architecture from `fixed_arc` and apply to model.

    Parameters
    ----------
    model : torch.nn.Module
        Model with mutables.
    fixed_arc : str or dict
        Path to the JSON that stores the architecture, or dict that stores the exported architecture.
    verbose : bool
        Print log messages if set to True

    Returns
    -------
    FixedArchitecture
        Mutator that is responsible for fixes the graph.
    """

    if isinstance(fixed_arc, str):
        with open(fixed_arc) as f:
            fixed_arc = json.load(f)
    architecture = CleanFixedArchitecture(model, fixed_arc, verbose)
    architecture.reset()

    # for the convenience of parameters counting
    architecture.replace_all_choice()
    return architecture



import os
import dgl.data.utils
import numpy as np
import scipy.io
import torch
from autogl.data import InMemoryStaticGraphSet
from .. import _dataset_registry


def get_binary_mask(total_size, indices):
    mask = torch.zeros(total_size)
    mask[indices] = 1
    return mask.bool()


@_dataset_registry.DatasetUniversalRegistry.register_dataset("hetero-acm-han")
class ACMHANDataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        data_path: str = os.path.join(path, 'raw', 'ACM.mat')
        _url: str = "https://data.dgl.ai/dataset/ACM.mat"
        if os.path.exists(data_path) and os.path.isfile(data_path):
            print(f"Using cached file {data_path}")
        else:
            dgl.data.utils.download(_url, data_path)
        data = scipy.io.loadmat(data_path)
        p_vs_l = data['PvsL']  # paper-field?
        p_vs_a = data['PvsA']  # paper-author
        p_vs_t = data['PvsT']  # paper-term, bag of words
        p_vs_c = data['PvsC']  # paper-conference, labels come from that

        # (1) KDD papers as class 0 (data mining),
        # (2) SIGMOD and VLDB papers as class 1 (database),
        # (3) SIGCOMM and MOBICOMM papers as class 2 (communication)
        conf_ids = [0, 1, 9, 10, 13]
        label_ids = [0, 1, 2, 2, 1]

        p_vs_c_filter = p_vs_c[:, conf_ids]
        p_selected = (p_vs_c_filter.sum(1) != 0).A1.nonzero()[0]
        p_vs_l = p_vs_l[p_selected]
        p_vs_a = p_vs_a[p_selected]
        p_vs_t = p_vs_t[p_selected]
        p_vs_c = p_vs_c[p_selected]

        hg = dgl.heterograph({
            ('paper', 'pa', 'author'): p_vs_a.nonzero(),
            ('author', 'ap', 'paper'): p_vs_a.transpose().nonzero(),
            ('paper', 'pf', 'field'): p_vs_l.nonzero(),
            ('field', 'fp', 'paper'): p_vs_l.transpose().nonzero()
        })
        hg.nodes['paper'].data['feat'] = torch.tensor(p_vs_t.toarray(), dtype=torch.float)

        pc_p, pc_c = p_vs_c.nonzero()
        labels = np.zeros(len(p_selected), dtype=np.int64)
        for conf_id, label_id in zip(conf_ids, label_ids):
            labels[pc_p[pc_c == conf_id]] = label_id
        hg.nodes['paper'].data['label'] = torch.LongTensor(labels)

        float_mask = np.zeros(len(pc_p))
        for conf_id in conf_ids:
            pc_c_mask = (pc_c == conf_id)
            float_mask[pc_c_mask] = np.random.permutation(np.linspace(0, 1, pc_c_mask.sum()))
        train_idx = np.where(float_mask <= 0.2)[0]
        val_idx = np.where((float_mask > 0.2) & (float_mask <= 0.3))[0]
        test_idx = np.where(float_mask > 0.3)[0]

        num_nodes = hg.number_of_nodes('paper')
        hg.nodes['paper'].data['train_mask'] = get_binary_mask(num_nodes, train_idx)
        hg.nodes['paper'].data['val_mask'] = get_binary_mask(num_nodes, val_idx)
        hg.nodes['paper'].data['test_mask'] = get_binary_mask(num_nodes, test_idx)

        super(ACMHANDataset, self).__init__([hg])
        self.schema.meta_paths = (('pa', 'ap'), ('pf', 'fp'))
        self.schema['target_node_type'] = 'paper'


@_dataset_registry.DatasetUniversalRegistry.register_dataset("hetero-acm-hgt")
class ACMHGTDataset(InMemoryStaticGraphSet):
    def __init__(self, path: str):
        data_path: str = os.path.join(path, 'raw', 'ACM.mat')
        _url: str = "https://data.dgl.ai/dataset/ACM.mat"
        if os.path.exists(data_path) and os.path.isfile(data_path):
            print(f"Using cached file {data_path}")
        else:
            dgl.data.utils.download(_url, data_path)
        data = scipy.io.loadmat(data_path)

        hg = dgl.heterograph({
            ('paper', 'written-by', 'author'): data['PvsA'].nonzero(),
            ('author', 'writing', 'paper'): data['PvsA'].transpose().nonzero(),
            ('paper', 'citing', 'paper'): data['PvsP'].nonzero(),
            ('paper', 'cited', 'paper'): data['PvsP'].transpose().nonzero(),
            ('paper', 'is-about', 'subject'): data['PvsL'].nonzero(),
            ('subject', 'has', 'paper'): data['PvsL'].transpose().nonzero(),
        })

        pvc = data['PvsC'].tocsr()
        p_selected = pvc.tocoo()
        # generate labels
        labels = pvc.indices
        hg.nodes['paper'].data['label'] = torch.tensor(labels).long()

        # generate train/val/test split
        pid = p_selected.row
        shuffle = np.random.permutation(pid)
        train_idx = torch.tensor(shuffle[0:800]).long()
        val_idx = torch.tensor(shuffle[800:900]).long()
        test_idx = torch.tensor(shuffle[900:]).long()
        num_nodes = hg.number_of_nodes('paper')
        hg.nodes['paper'].data['train_mask'] = get_binary_mask(num_nodes, train_idx)
        hg.nodes['paper'].data['val_mask'] = get_binary_mask(num_nodes, val_idx)
        hg.nodes['paper'].data['test_mask'] = get_binary_mask(num_nodes, test_idx)

        hg.node_dict = {}
        hg.edge_dict = {}
        for node_type in hg.ntypes:
            hg.node_dict[node_type] = len(hg.node_dict)
        for edge_type in hg.etypes:
            hg.edge_dict[edge_type] = len(hg.edge_dict)

        for edge_type in hg.etypes:
            hg.edges[edge_type].data['id'] = (
                    torch.ones(hg.number_of_edges(edge_type), dtype=torch.long) * len(hg.edge_dict)
            )

        # Random initialize input feature
        for node_type in hg.ntypes:
            embeddings = torch.Tensor(hg.number_of_nodes(node_type), 256)
            torch.nn.init.xavier_uniform_(embeddings)
            hg.nodes[node_type].data['feat'] = embeddings

        super(ACMHGTDataset, self).__init__([hg])
        self.schema['target_node_type'] = 'paper'



import torch.nn.functional
import typing as _typing
import dgl
from dgl.nn.pytorch.glob import (
    SumPooling, AvgPooling, MaxPooling, SortPooling
)
from .. import base_decoder, decoder_registry
from ...encoders import base_encoder
from ... import _utils


class _LogSoftmaxDecoder(torch.nn.Module):
    def forward(
            self, features: _typing.Sequence[torch.Tensor],
            _graph: dgl.DGLGraph, *__args, **__kwargs
    ):
        return torch.nn.functional.log_softmax(features[-1], dim=-1)


@decoder_registry.DecoderUniversalRegistry.register_decoder('log_softmax')
@decoder_registry.DecoderUniversalRegistry.register_decoder('log-softmax-decoder')
@decoder_registry.DecoderUniversalRegistry.register_decoder('log-softmax_decoder')
@decoder_registry.DecoderUniversalRegistry.register_decoder('log_softmax-decoder')
@decoder_registry.DecoderUniversalRegistry.register_decoder('log_softmax_decoder')
@decoder_registry.DecoderUniversalRegistry.register_decoder('LogSoftmax'.lower())
@decoder_registry.DecoderUniversalRegistry.register_decoder('LogSoftmax-decoder'.lower())
@decoder_registry.DecoderUniversalRegistry.register_decoder('LogSoftmax_decoder'.lower())
class LogSoftmaxDecoderMaintainer(base_decoder.BaseDecoderMaintainer):
    def _initialize(self, encoder, *args, **kwargs) -> _typing.Optional[bool]:
        self._decoder = _LogSoftmaxDecoder().to(self.device)
        return True


class _JKSumPoolDecoder(torch.nn.Module):
    def __init__(
            self, input_dimensions: _typing.Sequence[int],
            output_dimension: int, dropout: float,
            graph_pooling_type: str
    ):
        super(_JKSumPoolDecoder, self).__init__()
        self._linear_transforms: torch.nn.ModuleList = torch.nn.ModuleList()
        for input_dimension in input_dimensions:
            self._linear_transforms.append(
                torch.nn.Linear(input_dimension, output_dimension)
            )
        self._dropout: torch.nn.Dropout = torch.nn.Dropout(dropout)
        if not isinstance(graph_pooling_type, str):
            raise TypeError
        elif graph_pooling_type.lower() == 'sum':
            self.__pool = SumPooling()
        elif graph_pooling_type.lower() == 'mean':
            self.__pool = AvgPooling()
        elif graph_pooling_type.lower() == 'max':
            self.__pool = MaxPooling()
        else:
            raise NotImplementedError

    def forward(
            self, features: _typing.Sequence[torch.Tensor],
            graph: dgl.DGLGraph, *__args, **__kwargs
    ):
        if len(features) != len(self._linear_transforms):
            raise ValueError
        score_over_layer = 0
        for i, feature in enumerate(features):
            score_over_layer += self._dropout(self._linear_transforms[i](self.__pool(graph, feature)))
        return score_over_layer


@decoder_registry.DecoderUniversalRegistry.register_decoder('JKSumPoolMLP')
@decoder_registry.DecoderUniversalRegistry.register_decoder('JKSumPoolMLPDecoder')
@decoder_registry.DecoderUniversalRegistry.register_decoder('JKSumPoolMLP-decoder')
@decoder_registry.DecoderUniversalRegistry.register_decoder('JKSumPoolMLP-Decoder')
@decoder_registry.DecoderUniversalRegistry.register_decoder('JKSumPoolMLP_decoder')
@decoder_registry.DecoderUniversalRegistry.register_decoder('JKSumPoolMLP_Decoder')
@decoder_registry.DecoderUniversalRegistry.register_decoder('JKSumPoolMLP'.lower())
@decoder_registry.DecoderUniversalRegistry.register_decoder('JKSumPoolMLPDecoder'.lower())
@decoder_registry.DecoderUniversalRegistry.register_decoder('JKSumPoolMLP-Decoder'.lower())
@decoder_registry.DecoderUniversalRegistry.register_decoder('JKSumPoolMLP_Decoder'.lower())
class JKSumPoolDecoderMaintainer(base_decoder.BaseDecoderMaintainer):
    def _initialize(self, encoder: base_encoder.AutoHomogeneousEncoderMaintainer, *args, **kwargs) -> _typing.Optional[bool]:
        self._decoder = _JKSumPoolDecoder(
            list(encoder.get_output_dimensions()), self.output_dimension,
            self.hyper_parameters["dropout"],
            self.hyper_parameters["graph_pooling_type"]
        ).to(self.device)
        return True

    def __init__(
            self, output_dimension: _typing.Optional[int] = ...,
            device: _typing.Union[torch.device, str, int, None] = ...,
            *args, **kwargs
    ):
        super(JKSumPoolDecoderMaintainer, self).__init__(
            output_dimension, device, *args, **kwargs
        )
        self.hyper_parameter_space = (
            {
                "parameterName": "dropout",
                "type": "DOUBLE",
                "maxValue": 0.9,
                "minValue": 0.1,
                "scalingType": "LINEAR",
            },
            {
                "parameterName": "graph_pooling_type",
                "type": "CATEGORICAL",
                "feasiblePoints": ["sum", "mean", "max"],
            }
        )
        self.hyper_parameters = {
            "dropout": 0.5,
            "graph_pooling_type": "sum"
        }

class _SumPoolMLPDecoder(torch.nn.Module):
    def __init__(
            self, _final_dimension: int, hidden_dimension: int, output_dimension: int,
            _act: _typing.Optional[str], _dropout: _typing.Optional[float],
            num_graph_features: _typing.Optional[int]
    ):
        super(_SumPoolMLPDecoder, self).__init__()
        if (
                isinstance(num_graph_features, int)
                and num_graph_features > 0
        ):
            _final_dimension += num_graph_features
            self.__num_graph_features: _typing.Optional[int] = num_graph_features
        else:
            self.__num_graph_features: _typing.Optional[int] = None
        self._sumpool = SumPooling()
        self._fc1: torch.nn.Linear = torch.nn.Linear(
            _final_dimension, hidden_dimension
        )
        self._fc2: torch.nn.Linear = torch.nn.Linear(
            hidden_dimension, output_dimension
        )
        self._act: _typing.Optional[str] = _act
        self._dropout: _typing.Optional[float] = _dropout

    def forward(
            self, features: _typing.Sequence[torch.Tensor],
            data: dgl.DGLGraph, *__args, **__kwargs
    ):
        feature = features[-1]
        feature = self._sumpool(data, feature)
        if (
                isinstance(self.__num_graph_features, int)
                and self.__num_graph_features > 0
        ):
            if (
                    hasattr(data, 'gf') and
                    isinstance(data.gf, torch.Tensor) and data.gf.dim() == 2 and
                    data.gf.size() == (feature.size(0), self.__num_graph_features)
            ):
                graph_features: torch.Tensor = data.gf
            else:
                raise ValueError(
                    f"The provided data is expected to contain property 'gf' "
                    f"with {self.__num_graph_features} dimensions as graph feature"
                )
            feature: torch.Tensor = torch.cat([feature, graph_features], dim=-1)
        feature: torch.Tensor = self._fc1(feature)
        feature: torch.Tensor = _utils.activation.activation_func(feature, self._act)
        if isinstance(self._dropout, float) and 0 <= self._dropout <= 1:
            feature: torch.Tensor = torch.nn.functional.dropout(
                feature, self._dropout, self.training
            )
        feature: torch.Tensor = self._fc2(feature)
        return torch.nn.functional.log_softmax(feature, dim=-1)


@decoder_registry.DecoderUniversalRegistry.register_decoder('SumPoolMLP'.lower())
@decoder_registry.DecoderUniversalRegistry.register_decoder('SumPoolMLPDecoder'.lower())
@decoder_registry.DecoderUniversalRegistry.register_decoder('SumPoolMLP_Decoder'.lower())
class SumPoolMLPDecoderMaintainer(base_decoder.BaseDecoderMaintainer):
    def _initialize(self, encoder: base_encoder.AutoHomogeneousEncoderMaintainer, *args, **kwargs) -> _typing.Optional[bool]:
        if (
                isinstance(getattr(self, "num_graph_features"), int) and
                getattr(self, "num_graph_features") > 0
        ):
            num_graph_features: _typing.Optional[int] = getattr(self, "num_graph_features")
        else:
            num_graph_features: _typing.Optional[int] = None
        self._decoder = _SumPoolMLPDecoder(
            tuple(encoder.get_output_dimensions())[-1],
            self.hyper_parameters['hidden'], self.output_dimension,
            self.hyper_parameters['act'], self.hyper_parameters['dropout'],
            num_graph_features
        ).to(self.device)
        return True

    def __init__(
            self, output_dimension: _typing.Optional[int] = ...,
            device: _typing.Union[torch.device, str, int, None] = ...,
            *args, **kwargs
    ):
        super(SumPoolMLPDecoderMaintainer, self).__init__(
            output_dimension, device, *args, **kwargs
        )
        self.num_graph_features = kwargs.get("num_graph_features", 0)
        self.hyper_parameter_space = [
            {
                "parameterName": "hidden",
                "type": "INTEGER",
                "maxValue": 64,
                "minValue": 8,
                "scalingType": "LINEAR",
            },
            {
                "parameterName": "act",
                "type": "CATEGORICAL",
                "feasiblePoints": ["leaky_relu", "relu", "elu", "tanh"],
            },
            {
                "parameterName": "dropout",
                "type": "DOUBLE",
                "maxValue": 0.9,
                "minValue": 0.1,
                "scalingType": "LINEAR",
            }
        ]
        self.hyper_parameters = {
            "hidden": 32,
            "act": "relu",
            "dropout": 0.5
        }


class _TopKPoolDecoder(torch.nn.Module):
    def __init__(
            self, input_dimensions: _typing.Iterable[int],
            output_dimension: int, dropout: float
    ):
        super(_TopKPoolDecoder, self).__init__()
        k: int = min(len(list(input_dimensions)), 3)
        self.__pool: SortPooling = SortPooling(k)
        self.__linear_predictions: torch.nn.ModuleList = (
            torch.nn.ModuleList()
        )
        for layer, dimension in enumerate(input_dimensions):
            self.__linear_predictions.append(
                torch.nn.Linear(dimension * k, output_dimension)
            )
        self._dropout = torch.nn.Dropout(dropout)

    def forward(
            self, features: _typing.Sequence[torch.Tensor],
            graph: dgl.DGLGraph, *__args, **__kwargs
    ):
        cumulative_result = 0
        for i, h in enumerate(features):
            cumulative_result += self._dropout(self.__linear_predictions[i](self.__pool(graph, h)))
        return cumulative_result


@decoder_registry.DecoderUniversalRegistry.register_decoder('TopK')
@decoder_registry.DecoderUniversalRegistry.register_decoder('TopK-decoder')
@decoder_registry.DecoderUniversalRegistry.register_decoder('TopK-Decoder')
@decoder_registry.DecoderUniversalRegistry.register_decoder('TopK_decoder')
@decoder_registry.DecoderUniversalRegistry.register_decoder('TopK_Decoder')
@decoder_registry.DecoderUniversalRegistry.register_decoder('TopK'.lower())
@decoder_registry.DecoderUniversalRegistry.register_decoder('TopK-decoder'.lower())
@decoder_registry.DecoderUniversalRegistry.register_decoder('TopK_decoder'.lower())
class TopKDecoderMaintainer(base_decoder.BaseDecoderMaintainer):
    def _initialize(
            self, encoder: base_encoder.AutoHomogeneousEncoderMaintainer, *args, **kwargs
    ) -> _typing.Optional[bool]:
        self._decoder = _TopKPoolDecoder(
            encoder.get_output_dimensions(),
            self.output_dimension,
            self.hyper_parameters["dropout"]
        ).to(self.device)
        return True

    def __init__(
            self, output_dimension: _typing.Optional[int] = ...,
            device: _typing.Union[torch.device, str, int, None] = ...,
            *args, **kwargs
    ):
        super(TopKDecoderMaintainer, self).__init__(
            output_dimension, device, *args, **kwargs
        )
        self.hyper_parameter_space = [
            {
                "parameterName": "dropout",
                "type": "DOUBLE",
                "maxValue": 0.9,
                "minValue": 0.1,
                "scalingType": "LINEAR",
            }
        ]
        self.hyper_parameters = {
            "dropout": 0.5
        }


class _DotProductLinkPredictionDecoder(torch.nn.Module):
    def forward(
            self,
            features: _typing.Sequence[torch.Tensor],
            graph: dgl.DGLGraph,
            pos_edge: torch.Tensor,
            neg_edge: torch.Tensor,
            **__kwargs
    ):
        z = features[-1]
        edge_index = torch.cat([pos_edge, neg_edge], dim=-1)
        return (z[edge_index[0]] * z[edge_index[1]]).sum(dim=-1)


@decoder_registry.DecoderUniversalRegistry.register_decoder('LPDecoder'.lower())
@decoder_registry.DecoderUniversalRegistry.register_decoder('DOTProduct'.lower())
@decoder_registry.DecoderUniversalRegistry.register_decoder('DOTProductDecoder'.lower())
@decoder_registry.DecoderUniversalRegistry.register_decoder('lp-decoder'.lower())
@decoder_registry.DecoderUniversalRegistry.register_decoder('dot-product'.lower())
class DotProductLinkPredictionDecoderMaintainer(base_decoder.BaseDecoderMaintainer):
    def _initialize(self, *args, **kwargs):
        self._decoder = _DotProductLinkPredictionDecoder()



"""
auto graph model
a list of models with their hyper parameters
NOTE: neural architecture search (NAS) maybe included here
"""
import copy
import logging
import typing as _typing
import torch
import torch.nn.functional as F
from copy import deepcopy
from autogl.module.hpo.auto_module import AutoModule

base_approach_logger: logging.Logger = logging.getLogger("BaseModel")


def activate_func(x, func):
    if func == "tanh":
        return torch.tanh(x)
    elif hasattr(F, func):
        return getattr(F, func)(x)
    elif func == "":
        pass
    else:
        raise TypeError("PyTorch does not support activation function {}".format(func))

    return x


class BaseAutoModel(AutoModule):
    def __init__(self, input_dimension, output_dimension, device, **kwargs):
        self._input_dimension = input_dimension
        self._output_dimension = output_dimension
        self._model = None
        self._kwargs = kwargs
        super(BaseAutoModel, self).__init__(device)

    def to(self, device):
        # compatible with v0.2
        self.to_device(device)

    def to_device(self, device):
        self.device = device
        if self.model is not None:
            self.model.to(self.device)

    @property
    def model(self):
        return self._model

    def from_hyper_parameter(self, hp, **kwargs):
        kw = deepcopy(self._kwargs)
        kw.update(kwargs)
        ret_self = self.__class__(
            self.input_dimension,
            self.output_dimension,
            self.device,
            **kw
        )
        hp_now = dict(self.hyper_parameters)
        hp_now.update(hp)
        ret_self.hyper_parameters = hp_now
        ret_self.initialize()
        return ret_self

    @property
    def input_dimension(self):
        return self._input_dimension
    
    @input_dimension.setter
    def input_dimension(self, input_dimension):
        self._input_dimension = input_dimension

    @property
    def output_dimension(self):
        return self._output_dimension
    
    @output_dimension.setter
    def output_dimension(self, output_dimension):
        self._output_dimension = output_dimension


class _BaseBaseModel:
    # todo: after renaming the experimental base class _BaseModel to BaseModel,
    #       rename this class to _BaseModel
    """
    The base class for class BaseModel,
    designed to implement some basic functionality of BaseModel.
    --  Designed by ZiXin Sun
    """

    @classmethod
    def __formulate_device(
        cls, device: _typing.Union[str, torch.device] = ...
    ) -> torch.device:
        if type(device) == torch.device or (
            type(device) == str and device.strip().lower() != "auto"
        ):
            return torch.device(device)
        elif torch.cuda.is_available() and torch.cuda.device_count() > 0:
            return torch.device("cuda")
        else:
            return torch.device("cpu")

    @property
    def device(self) -> torch.device:
        return self.__device

    @device.setter
    def device(self, __device: _typing.Union[str, torch.device, None]):
        self.__device: torch.device = self.__formulate_device(__device)

    @property
    def model(self) -> _typing.Optional[torch.nn.Module]:
        if self._model is None:
            base_approach_logger.debug(
                "property of model NOT initialized before accessing"
            )
        return self._model

    @model.setter
    def model(self, _model: torch.nn.Module) -> None:
        if not isinstance(_model, torch.nn.Module):
            raise TypeError(
                "the property of model MUST be an instance of " "torch.nn.Module"
            )
        self._model = _model

    def _initialize(self):
        raise NotImplementedError

    def initialize(self) -> bool:
        """
        Initialize the model in case that the model has NOT been initialized
        :return: whether self._initialize() method called
        """
        if not self.__is_initialized:
            self._initialize()
            self.__is_initialized = True
            return True
        return False

    # def to(self, *args, **kwargs):
    #     """
    #     Due to the signature of to() method in class BaseApproach
    #     is inconsistent with the signature of the method
    #     in the base class torch.nn.Module,
    #     this intermediate overridden method is necessary to
    #     walk around (bypass) the inspection for
    #     signature of overriding method.
    #     :param args: positional arguments list
    #     :param kwargs: keyword arguments dict
    #     :return: self
    #     """
    #     return super(_BaseBaseModel, self).to(*args, **kwargs)

    def forward(self, *args, **kwargs):
        if self.model is not None and isinstance(self.model, torch.nn.Module):
            return self.model(*args, **kwargs)
        else:
            raise NotImplementedError

    def __init__(
        self,
        model: _typing.Optional[torch.nn.Module] = None,
        initialize: bool = False,
        device: _typing.Union[str, torch.device] = ...,
    ):
        if type(initialize) != bool:
            raise TypeError
        super(_BaseBaseModel, self).__init__()
        self.__device: torch.device = self.__formulate_device(device)
        self._model: _typing.Optional[torch.nn.Module] = model
        self.__is_initialized: bool = False
        if initialize:
            self.initialize()


class _BaseModel(_BaseBaseModel, BaseAutoModel):
    """
    The upcoming root base class for Model, i.e. BaseModel
    --  Designed by ZiXin Sun
    """

    # todo: Deprecate and remove the legacy class "BaseModel",
    #       then rename this class to "BaseModel",
    #       correspondingly, this class will no longer extend
    #       the legacy class "BaseModel" after the removal.
    def _initialize(self):
        raise NotImplementedError

    def to(self, device: torch.device):
        self.device = device
        if self.model is not None and isinstance(self.model, torch.nn.Module):
            self.model.to(self.device)
        return super().to(device)

    @property
    def space(self) -> _typing.Sequence[_typing.Dict[str, _typing.Any]]:
        # todo: deprecate and remove in future major version
        return self.__hyper_parameter_space

    @property
    def hyper_parameter_space(self):
        return self.__hyper_parameter_space

    @hyper_parameter_space.setter
    def hyper_parameter_space(
        self, space: _typing.Sequence[_typing.Dict[str, _typing.Any]]
    ):
        self.__hyper_parameter_space = space

    @property
    def hyper_parameter(self) -> _typing.Dict[str, _typing.Any]:
        return self.__hyper_parameter

    @hyper_parameter.setter
    def hyper_parameter(self, _hyper_parameter: _typing.Dict[str, _typing.Any]):
        if not isinstance(_hyper_parameter, dict):
            raise TypeError
        self.__hyper_parameter = _hyper_parameter

    def get_hyper_parameter(self) -> _typing.Dict[str, _typing.Any]:
        """
        todo: consider deprecating this trivial getter method in the future
        :return: copied hyper parameter
        """
        return copy.deepcopy(self.__hyper_parameter)

    def __init__(
        self,
        model: _typing.Optional[torch.nn.Module] = None,
        initialize: bool = False,
        hyper_parameter_space: _typing.Sequence[_typing.Any] = ...,
        hyper_parameter: _typing.Dict[str, _typing.Any] = ...,
        device: _typing.Union[str, torch.device] = ...,
    ):
        if type(initialize) != bool:
            raise TypeError
        super(_BaseModel, self).__init__(model, initialize, device)
        if hyper_parameter_space != Ellipsis and isinstance(
            hyper_parameter_space, _typing.Sequence
        ):
            self.__hyper_parameter_space: _typing.Sequence[
                _typing.Dict[str, _typing.Any]
            ] = hyper_parameter_space
        else:
            self.__hyper_parameter_space: _typing.Sequence[
                _typing.Dict[str, _typing.Any]
            ] = []
        if hyper_parameter != Ellipsis and isinstance(hyper_parameter, dict):
            self.__hyper_parameter: _typing.Dict[str, _typing.Any] = hyper_parameter
        else:
            self.__hyper_parameter: _typing.Dict[str, _typing.Any] = {}

    def from_hyper_parameter(self, hyper_parameter: _typing.Dict[str, _typing.Any]):
        raise NotImplementedError


class ClassificationModel(_BaseModel):
    def _initialize(self):
        raise NotImplementedError

    def from_hyper_parameter(
        self, hyper_parameter: _typing.Dict[str, _typing.Any]
    ) -> "ClassificationModel":
        new_model: ClassificationModel = self.__class__(
            num_features=self.num_features,
            num_classes=self.num_classes,
            device=self.device,
            init=False,
        )
        _hyper_parameter = self.hyper_parameter
        _hyper_parameter.update(hyper_parameter)
        new_model.hyper_parameter = _hyper_parameter
        new_model.initialize()
        return new_model

    def __init__(
        self,
        num_features: int = ...,
        num_classes: int = ...,
        num_graph_features: int = ...,
        device: _typing.Union[str, torch.device] = ...,
        hyper_parameter_space: _typing.Sequence[_typing.Any] = ...,
        hyper_parameter: _typing.Dict[str, _typing.Any] = ...,
        init: bool = False,
        **kwargs
    ):
        if "initialize" in kwargs:
            del kwargs["initialize"]
        super(ClassificationModel, self).__init__(
            initialize=init,
            hyper_parameter_space=hyper_parameter_space,
            hyper_parameter=hyper_parameter,
            device=device,
            **kwargs
        )
        if num_classes != Ellipsis and type(num_classes) == int:
            self.__num_classes: int = num_classes if num_classes > 0 else 0
        else:
            self.__num_classes: int = 0
        if num_features != Ellipsis and type(num_features) == int:
            self.__num_features: int = num_features if num_features > 0 else 0
        else:
            self.__num_features: int = 0
        if num_graph_features != Ellipsis and type(num_graph_features) == int:
            if num_graph_features > 0:
                self.__num_graph_features: int = num_graph_features
            else:
                self.__num_graph_features: int = 0
        else:
            self.__num_graph_features: int = 0

    def __repr__(self) -> str:
        import yaml

        return yaml.dump(self.hyper_parameter)

    @property
    def num_classes(self) -> int:
        return self.__num_classes

    @num_classes.setter
    def num_classes(self, __num_classes: int):
        if type(__num_classes) != int:
            raise TypeError
        if not __num_classes > 0:
            raise ValueError
        self.__num_classes = __num_classes if __num_classes > 0 else 0

    @property
    def num_features(self) -> int:
        return self.__num_features

    @num_features.setter
    def num_features(self, __num_features: int):
        if type(__num_features) != int:
            raise TypeError
        if not __num_features > 0:
            raise ValueError
        self.__num_features = __num_features if __num_features > 0 else 0

    def get_num_classes(self) -> int:
        # todo: consider replacing with property with getter and setter
        return self.__num_classes

    def set_num_classes(self, num_classes: int) -> None:
        # todo: consider replacing with property with getter and setter
        if type(num_classes) != int:
            raise TypeError
        self.__num_classes = num_classes if num_classes > 0 else 0

    def get_num_features(self) -> int:
        # todo: consider replacing with property with getter and setter
        return self.__num_features

    def set_num_features(self, num_features: int):
        # todo: consider replacing with property with getter and setter
        if type(num_features) != int:
            raise TypeError
        self.__num_features = num_features if num_features > 0 else 0

    def set_num_graph_features(self, num_graph_features: int):
        # todo: consider replacing with property with getter and setter
        if type(num_graph_features) != int:
            raise TypeError
        else:
            if num_graph_features > 0:
                self.__num_graph_features = num_graph_features
            else:
                self.__num_graph_features = 0


class _ClassificationModel(torch.nn.Module):
    def __init__(self):
        super(_ClassificationModel, self).__init__()

    def cls_encode(self, data) -> torch.Tensor:
        raise NotImplementedError

    def cls_decode(self, x: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError

    def cls_forward(self, data) -> torch.Tensor:
        return self.cls_decode(self.cls_encode(data))


class ClassificationSupportedSequentialModel(_ClassificationModel):
    def __init__(self):
        super(ClassificationSupportedSequentialModel, self).__init__()

    @property
    def sequential_encoding_layers(self) -> torch.nn.ModuleList:
        raise NotImplementedError

    def cls_encode(self, data) -> torch.Tensor:
        raise NotImplementedError

    def cls_decode(self, x: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError



"""
Solver base class

Provide some standard solver interface.
"""

from typing import Any, Iterable, Tuple
from copy import deepcopy

import torch

from ..module.feature import FEATURE_DICT
from ..module.hpo import HPO_DICT
from ..module.model import EncoderUniversalRegistry, DecoderUniversalRegistry, ModelUniversalRegistry
from ..module.nas.algorithm import NAS_ALGO_DICT
from ..module.nas.estimator import NAS_ESTIMATOR_DICT
from ..module.nas.space import NAS_SPACE_DICT
from ..module import BaseFeatureEngineer, BaseHPOptimizer, BaseTrainer
from .utils import LeaderBoard
from ..utils import get_logger

LOGGER = get_logger("BaseSolver")


def _initialize_single_model(model):
    encoder, decoder = None, None
    if "encoder" in model:
        # initialize encoder
        name = model["encoder"].pop("name")
        encoder = EncoderUniversalRegistry.get_encoder(name)(**model["encoder"])
    if "decoder" in model:
        # initialize decoder
        name = model["decoder"].pop("name")
        decoder = DecoderUniversalRegistry.get_decoder(name)(**model["decoder"])
        return (encoder, decoder)

    if "name" in model:
        # whole model
        name = model.pop("name")
        encoder = ModelUniversalRegistry.get_model(name)(**model)
    return encoder

def _parse_hp_space(spaces):
    if spaces is None:
        return None
    for space in spaces:
        if "cutFunc" in space and isinstance(space["cutFunc"], str):
            space["cutFunc"] = eval(space["cutFunc"])
    return spaces

def _parse_model_hp(model):
    assert isinstance(model, dict)
    if "encoder" in model and "decoder" in model:
        return {
            "encoder": _parse_hp_space(model["encoder"].pop("hp_space", None)),
            "decoder": _parse_hp_space(model["decoder"].pop("hp_space", None)),
        }
    elif "encoder" in model:
        return {
            "encoder": _parse_hp_space(model["encoder"].pop("hp_space", None)),
            "decoder": None,
        }
    else:
        return _parse_hp_space(model.pop("hp_space", None))

class BaseSolver:
    r"""
    Base solver class, define some standard solver interfaces.

    Parameters
    ----------
    feature_module: autogl.module.feature.BaseFeatureEngineer or str or None
        The (name of) auto feature engineer used to process the given dataset.
        Disable feature engineer by setting it to ``None``.

    graph_models: list of autogl.module.model.BaseModel or str
        The (name of) models to be optimized as backbone.

    hpo_module: autogl.module.hpo.BaseHPOptimizer or str or None
        The (name of) hpo module used to search for best hyper parameters.
        Disable hpo by setting it to ``None``.

    ensemble_module: autogl.module.ensemble.BaseEnsembler or str or None
        The (name of) ensemble module used to ensemble the multi-models found.
        Disable ensemble by setting it to ``None``.

    max_evals: int (Optional)
        If given, will set the number eval times the hpo module will use.
        Only be effective when hpo_module is  of type ``str``. Default ``50``.

    default_trainer: str or list of str (Optional)
        Default trainer class to be used.
        If a single trainer class is given, will set all trainer to default trainer.
        If a list of trainer class is given, will set every model with corresponding trainer
        cls. Default ``None``.

    trainer_hp_space: list of dict (Optional)
        trainer hp space or list of trainer hp spaces configuration.
        If a single trainer hp is given, will specify the hp space of trainer for every model.
        If a list of trainer hp is given, will specify every model with corrsponding
        trainer hp space. Default ``None``

    model_hp_spaces: list of list of dict (Optional)
        model hp space configuration.
        If given, will specify every hp space of every passed model. Default ``None``.

    size: int (Optional)
        The max models ensemble module will use. Default ``None``.

    device: torch.device or str
        The device where model will be running on. If set to ``auto``, will use gpu when available.
        You can also specify the device by directly giving ``gpu`` or ``cuda:0``, etc.
        Default ``auto``.
    """

    # pylint: disable=W0102

    def __init__(
        self,
        feature_module,
        graph_models,
        nas_spaces,
        nas_algorithms,
        nas_estimators,
        hpo_module,
        ensemble_module,
        max_evals=50,
        default_trainer=None,
        trainer_hp_space=None,
        model_hp_spaces=None,
        size=4,
        device="auto",
    ):

        # set default device
        if device == "auto":
            self.runtime_device = torch.device(
                "cuda" if torch.cuda.is_available() else "cpu"
            )
        elif isinstance(device, torch.device):
            self.runtime_device = device
        elif isinstance(device, str) and (device == "cpu" or device.startswith("cuda")):
            self.runtime_device = torch.device(device)
        else:
            LOGGER.error("Cannot parse device %s", str(device))
            raise ValueError("Cannot parse device {}".format(device))

        # initialize modules
        self.graph_model_list = []
        self.set_graph_models(
            graph_models, default_trainer, trainer_hp_space, model_hp_spaces
        )
        self.set_feature_module(feature_module)
        self.set_hpo_module(hpo_module, max_evals=max_evals)
        self.set_ensemble_module(ensemble_module, size=size)
        self.set_nas_module(nas_algorithms, nas_spaces, nas_estimators)

        # initialize leaderboard
        self.leaderboard = None

        # trained model is saved here
        self.trained_models = {}

    def set_feature_module(
        self,
        feature_module,
        *args,
        **kwargs,
    ) -> "BaseSolver":
        r"""
        Set the feature module of current solver.

        Parameters
        ----------
        feature_module: autogl.module.feature.BaseFeature or str or None
            The (name of) auto feature engineer used to process the given dataset.
            Disable feature engineer by setting it to ``None``.

        Returns
        -------
        self: autogl.solver.BaseSolver
            A reference of current solver.
        """
        # load feature engineer module

        def get_feature(feature_engineer):
            if isinstance(feature_engineer, BaseFeatureEngineer):
                return feature_engineer
            if isinstance(feature_engineer, str):
                if feature_engineer in FEATURE_DICT:
                    return FEATURE_DICT[feature_engineer](*args, **kwargs)
                raise ValueError(
                    "cannot find feature engineer %s." % (feature_engineer)
                )
            raise TypeError(
                f"Cannot parse feature argument {str(feature_engineer)} of"
                f" type {str(type(feature_engineer))}"
            )

        if feature_module is None:
            self.feature_module = None
        elif isinstance(feature_module, (BaseFeatureEngineer, str)):
            self.feature_module = get_feature(feature_module)
        elif isinstance(feature_module, Iterable):
            self.feature_module = get_feature(feature_module[0])
            for feature_engineer in feature_module[1:]:
                self.feature_module &= get_feature(feature_engineer)
        else:
            raise ValueError(
                "need feature module to be str or a BaseFeatureEngineer instance, get",
                type(feature_module),
                "instead.",
            )

        return self

    def set_graph_models(
        self,
        graph_models,
        default_trainer=None,
        trainer_hp_space=None,
        model_hp_spaces=None,
    ) -> "BaseSolver":
        r"""
        Set the graph models used in current solver.

        Parameters
        ----------
        graph_models: list of autogl.module.model.BaseModel or list of str
            The (name of) models to be optimized as backbone.

        default_trainer: str or list of str (Optional)
            Default trainer class to be used.
            If a single trainer class is given, will set all trainer to default trainer.
            If a list of trainer class is given, will set every model with corresponding trainer
            cls. Default ``None``.

        trainer_hp_space: list of dict (Optional)
            trainer hp space or list of trainer hp spaces configuration.
            If a single trainer hp is given, will specify the hp space of trainer for every model.
            If a list of trainer hp is given, will specify every model with corrsponding
            trainer hp space.
            Default ``None``.

        model_hp_spaces: list of list of dict (Optional)
            model hp space configuration.
            If given, will specify every hp space of every passed model. Default ``None``.

        Returns
        -------
        self: autogl.solver.BaseSolver
            A reference of current solver.
        """
        self.gml = graph_models
        self._default_trainer = default_trainer
        self._trainer_hp_space = trainer_hp_space
        self._model_hp_spaces = model_hp_spaces
        return self

    def set_hpo_module(self, hpo_module, *args, **kwargs) -> "BaseSolver":
        r"""
        Set the hpo module used in current solver.

        Parameters
        ----------
        hpo_module: autogl.module.hpo.BaseHPOptimizer or str or None
            The (name of) hpo module used to search for best hyper parameters.
            Disable hpo by setting it to ``None``.

        Returns
        -------
        self: autogl.solver.BaseSolver
            A reference of current solver.
        """

        # load hpo module
        if hpo_module is None:
            self.hpo_module = None
        elif isinstance(hpo_module, BaseHPOptimizer):
            self.hpo_module = hpo_module
        elif isinstance(hpo_module, str):
            if hpo_module in HPO_DICT:
                self.hpo_module = HPO_DICT[hpo_module](*args, **kwargs)
            else:
                raise KeyError("cannot find hpo module %s." % (hpo_module))
        else:
            raise ValueError(
                "need hpo module to be str or a BaseHPOtimizer instance, get",
                type(hpo_module),
                "instead.",
            )
        return self

    def set_nas_module(
        self, nas_algorithms=None, nas_spaces=None, nas_estimators=None
    ) -> "BaseSolver":
        """
        Set the neural architecture search module in current solver.

        Parameters
        ----------
        nas_spaces: (list of) `autogl.module.hpo.nas.GraphSpace`
            The search space of nas. You can pass a list of space to enable
            multiple space search. If list passed, the length of `nas_spaces`,
            `nas_algorithms` and `nas_estimators` should be the same. If set
            to `None`, will disable the whole nas module.

        nas_algorithms: (list of) `autogl.module.hpo.nas.BaseNAS`
            The search algorithm of nas. You can pass a list of algorithms
            to enable multiple algorithms search. If list passed, the length of
            `nas_spaces`, `nas_algorithms` and `nas_estimators` should be the same.
            Default `None`.

        nas_estimators: (list of) `autogl.module.hpo.nas.BaseEstimators`
            The nas estimators. You can pass a list of estimators to enable multiple
            estimators search. If list passed, the length of `nas_spaces`, `nas_algorithms`
            and `nas_estimators` should be the same. Default `None`.

        Returns
        -------
        self: autogl.solver.BaseSolver
            A reference of current solver.
        """
        if nas_algorithms is None and nas_estimators is None and nas_spaces is None:
            self.nas_algorithms = self.nas_estimators = self.nas_spaces = None
            return
        assert None not in [
            nas_algorithms,
            nas_estimators,
            nas_spaces,
        ], "The algorithms, estimators and spaces should all be set"

        nas_algorithms = (
            nas_algorithms
            if isinstance(nas_algorithms, Iterable)
            else [nas_algorithms]
        )
        nas_spaces = (
            nas_spaces if isinstance(nas_spaces, Iterable) else [nas_spaces]
        )
        nas_estimators = (
            nas_estimators
            if isinstance(nas_estimators, Iterable)
            else [nas_estimators]
        )

        # parse all str elements
        nas_algorithms = [
            algo if not isinstance(algo, str) else NAS_ALGO_DICT[algo]()
            for algo in nas_algorithms
        ]
        nas_spaces = [
            space if not isinstance(space, str) else NAS_SPACE_DICT[space]()
            for space in nas_spaces
        ]
        nas_estimators = [
            estimator
            if not isinstance(estimator, str)
            else NAS_ESTIMATOR_DICT[estimator]()
            for estimator in nas_estimators
        ]

        max_number = max([len(x) for x in [nas_algorithms, nas_spaces, nas_estimators]])
        assert all(
            [
                len(x) in [1, max_number]
                for x in [nas_algorithms, nas_spaces, nas_estimators]
            ]
        ), "lengths of algorithms/spaces/estimators do not match!"

        self.nas_algorithms = (
            [deepcopy(nas_algorithms) for _ in range(max_number)]
            if len(nas_algorithms) == 1 and max_number > 1
            else nas_algorithms
        )
        self.nas_spaces = (
            [deepcopy(nas_spaces) for _ in range(max_number)]
            if len(nas_spaces) == 1 and max_number > 1
            else nas_spaces
        )
        self.nas_estimators = (
            [deepcopy(nas_estimators) for _ in range(max_number)]
            if len(nas_estimators) == 1 and max_number > 1
            else nas_estimators
        )

        return self

    def set_ensemble_module(self, ensemble_module, *args, **kwargs) -> "BaseSolver":
        r"""
        Set the ensemble module used in current solver.

        Parameters
        ----------
        ensemble_module: autogl.module.ensemble.BaseEnsembler or str or None
            The (name of) ensemble module used to ensemble the multi-models found.
            Disable ensemble by setting it to ``None``.

        Returns
        -------
        self: autogl.solver.BaseSolver
            A reference of current solver.
        """

        raise NotImplementedError()

    def fit(self, *args, **kwargs) -> "BaseSolver":
        r"""
        Fit current solver on given dataset.

        Returns
        -------
        self: autogl.solver.BaseSolver
            A reference of current solver.
        """
        raise NotImplementedError()

    def fit_predict(self, *args, **kwargs) -> Any:
        r"""
        Fit current solver on given dataset and return the predicted value.

        Returns
        -------
        result: Any
            The predicted result
        """
        raise NotImplementedError()

    def predict(self, *args, **kwargs) -> Any:
        r"""
        Predict the node class number.

        Returns
        -------
        result: Any
            The predicted result
        """
        raise NotImplementedError()

    def get_leaderboard(self) -> LeaderBoard:
        r"""
        Get the current leaderboard of this solver.

        Returns
        -------
        lb: autogl.solver.leaderboard
            A leaderboard instance.
        """
        return self.leaderboard

    def get_model_by_name(self, name) -> BaseTrainer:
        r"""
        Find and get the model instance by name.

        Parameters
        ----------
        name: str
            The name of model

        Returns
        -------
        trainer: autogl.module.train.BaseTrainer
            A trainer instance containing the trained models and training status.
        """
        assert name in self.trained_models, "cannot find model by name" + name
        return self.trained_models[name]

    def get_model_by_performance(self, index) -> Tuple[BaseTrainer, str]:
        r"""
        Find and get the model instance by performance.

        Parameters
        ----------
        index: int
            The performance index of model (from good to bad). Index from 0.

        Returns
        -------
        trainer: autogl.module.train.BaseTrainer
            A trainer instance containing the trained models and training status.
        name: str
            The name of current trainer.
        """
        name = self.leaderboard.get_best_model(index=index)
        return self.trained_models[name], name

    @classmethod
    def from_config(cls, path_or_dict, filetype="auto") -> "BaseSolver":
        r"""
        Load solver from config file.

        You can use this function to directly load a solver from predefined config dict
        or config file path.

        Parameters
        ----------
        path_or_dict: str or dict
            The path to the config file or the config dictionary object

        filetype: str
            The filetype the given file if the path is specified.

        Returns
        -------
        solver: autogl.solver.AutoGraphClassifier
            The solver that is created from given file or dictionary.
        """
        raise NotImplementedError()



import typing as _typing
import networkx as nx
from autogl.data.graph._general_static_graph import GeneralStaticGraph


class HomogeneousStaticGraphToNetworkX:
    def __init__(
            self, remove_self_loops: bool = False, to_undirected: bool = False,
            *__args, **__kwargs
    ):
        if not isinstance(remove_self_loops, bool):
            raise TypeError
        if not isinstance(to_undirected, bool):
            raise TypeError
        self.__remove_self_loops: bool = remove_self_loops
        self.__to_undirected: bool = to_undirected

    def __call__(
            self, homogeneous_static_graph: GeneralStaticGraph,
            remove_self_loops: _typing.Optional[bool] = ...,
            to_undirected: _typing.Optional[bool] = ...,
            *args, **kwargs
    ):
        if not isinstance(homogeneous_static_graph, GeneralStaticGraph):
            raise TypeError
        elif not (
                homogeneous_static_graph.nodes.is_homogeneous and
                homogeneous_static_graph.edges.is_homogeneous
        ):
            raise ValueError("Only homogeneous static graph can be converted to NetworkX")

        if not (remove_self_loops in (Ellipsis, None) or isinstance(remove_self_loops, bool)):
            raise TypeError
        else:
            __remove_self_loops: bool = (
                remove_self_loops if isinstance(remove_self_loops, bool)
                else self.__remove_self_loops
            )
        if not (to_undirected in (Ellipsis, None) or isinstance(to_undirected, bool)):
            raise TypeError
        else:
            __to_undirected: bool = (
                to_undirected if isinstance(to_undirected, bool)
                else self.__to_undirected
            )

        num_nodes: int = homogeneous_static_graph.edges.connections.max().item() + 1
        # todo: Note that this is an assumption

        g: nx.Graph = nx.Graph() if __to_undirected else nx.DiGraph()
        g.add_nodes_from(range(num_nodes))

        nodes_data: _typing.MutableMapping[str, list] = {}
        for data_key in homogeneous_static_graph.nodes.data:
            nodes_data[data_key] = (
                homogeneous_static_graph.nodes.data[data_key].squeeze().tolist()
            )
        edges_data: _typing.MutableMapping[str, list] = {}
        for data_key in homogeneous_static_graph.edges.data:
            edges_data[data_key] = (
                homogeneous_static_graph.edges.data[data_key].squeeze().tolist()
            )
        for i, (u, v) in enumerate(homogeneous_static_graph.edges.connections.t().tolist()):
            if __remove_self_loops and v == u:
                continue
            g.add_edge(u, v)
            for data_key in edges_data:
                g[u][v][data_key] = edges_data[data_key][i]
        for data_key in nodes_data:
            for i, feature_dict in g.nodes(data=True):
                feature_dict.update(
                    {data_key: nodes_data[data_key][i]}
                )
        return g



from ....backend import DependentBackend

if DependentBackend.is_pyg():
    from .graphcl import GraphCLSemisupervisedTrainer, GraphCLUnsupervisedTrainer, BaseContrastiveTrainer
    from .utils import get_view_by_name
    __all__ = [
        "GraphCLSemisupervisedTrainer",
        "GraphCLUnsupervisedTrainer",
        "BaseContrastiveTrainer",
        "get_view_by_name"
    ]
else:
    from .graphcl import BaseContrastiveTrainer
    __all__ = [
        "BaseContrastiveTrainer"
    ]



"""
ensemble module
"""


class BaseEnsembler:
    def __init__(self, *args, **kwargs):
        super().__init__()

    def fit(self, predictions, label, identifiers, feval, *args, **kwargs):
        pass

    def ensemble(self, predictions, identifiers, *args, **kwargs):
        pass

    @classmethod
    def build_ensembler_from_args(cls, args):
        """Build a new ensembler instance."""
        raise NotImplementedError(
            "Ensembler must implement the build_ensembler_from_args method"
        )



import typing as _typing
from autogl.utils import universal_registry
from . import base_decoder


class DecoderUniversalRegistry(universal_registry.UniversalRegistryBase):
    @classmethod
    def register_decoder(cls, name: str) -> _typing.Callable[
        [_typing.Type[base_decoder.BaseDecoderMaintainer]],
        _typing.Type[base_decoder.BaseDecoderMaintainer]
    ]:
        def register_decoder(
                _decoder: _typing.Type[base_decoder.BaseDecoderMaintainer]
        ) -> _typing.Type[base_decoder.BaseDecoderMaintainer]:
            if not issubclass(_decoder, base_decoder.BaseDecoderMaintainer):
                raise TypeError
            else:
                cls[name] = _decoder
                return _decoder

        return register_decoder

    @classmethod
    def get_decoder(cls, name: str) -> _typing.Type[base_decoder.BaseDecoderMaintainer]:
        if name not in cls:
            raise ValueError(f"Decoder with name \"{name}\" not exist")
        else:
            return cls[name]



from autogl import backend as _backend
from ._dataset_registry import (
    DatasetUniversalRegistry,
    build_dataset_from_name
)

from ._gtn_data import (
    GTNACMDataset,
    GTNDBLPDataset,
    GTNIMDBDataset,
)

from ._matlab_matrix import BlogCatalogDataset, WIKIPEDIADataset
from ._ogb import (
    OGBNProductsDataset, OGBNProteinsDataset, OGBNArxivDataset, OGBNPapers100MDataset,
    OGBLPPADataset, OGBLCOLLABDataset, OGBLDDIDataset, OGBLCitation2Dataset,
    OGBGMOLHIVDataset, OGBGMOLPCBADataset, OGBGPPADataset, OGBGCode2Dataset
)

if _backend.DependentBackend.is_dgl():
    from ._dgl import (
        CoraDataset,
        CiteSeerDataset,
        PubMedDataset,
        RedditDataset,
        AmazonComputersDataset,
        AmazonPhotoDataset,
        CoauthorPhysicsDataset,
        CoauthorCSDataset,
        MUTAGDataset,
        ENZYMESDataset,
        IMDBBinaryDataset,
        IMDBMultiDataset,
        RedditBinaryDataset,
        REDDITMulti5KDataset,
        COLLABDataset,
        ProteinsDataset,
        PTCMRDataset,
        NCI1Dataset
    )
    from ._heterogeneous_datasets import ACMHANDataset, ACMHGTDataset

elif _backend.DependentBackend.is_pyg():
    from ._pyg import (
        CoraDataset,
        CiteSeerDataset,
        PubMedDataset,
        FlickrDataset,
        RedditDataset,
        AmazonComputersDataset,
        AmazonPhotoDataset,
        CoauthorPhysicsDataset,
        CoauthorCSDataset,
        PPIDataset,
        QM9Dataset,
        MUTAGDataset,
        ENZYMESDataset,
        IMDBBinaryDataset,
        IMDBMultiDataset,
        RedditBinaryDataset,
        REDDITMulti5KDataset,
        REDDITMulti12KDataset,
        COLLABDataset,
        ProteinsDataset,
        PTCMRDataset,
        NCI1Dataset,
        NCI109Dataset,
        ModelNet10TrainingDataset,
        ModelNet10TestDataset,
        ModelNet40TrainingDataset,
        ModelNet40TestDataset
    )

if _backend.DependentBackend.is_pyg():
    __all__ = [
        "CoraDataset",
        "CiteSeerDataset",
        "PubMedDataset",
        "FlickrDataset",
        "RedditDataset",
        "AmazonComputersDataset",
        "AmazonPhotoDataset",
        "CoauthorPhysicsDataset",
        "CoauthorCSDataset",
        "PPIDataset",
        "QM9Dataset",
        "MUTAGDataset",
        "ENZYMESDataset",
        "IMDBBinaryDataset",
        "IMDBMultiDataset",
        "RedditBinaryDataset",
        "REDDITMulti5KDataset",
        "REDDITMulti12KDataset",
        "COLLABDataset",
        "ProteinsDataset",
        "PTCMRDataset",
        "NCI1Dataset",
        "NCI109Dataset",
        "ModelNet10TrainingDataset",
        "ModelNet10TestDataset",
        "ModelNet40TrainingDataset",
        "ModelNet40TestDataset",
        "OGBNProductsDataset",
        "OGBNProteinsDataset",
        "OGBNArxivDataset",
        "OGBNPapers100MDataset",
        "OGBLPPADataset",
        "OGBLCOLLABDataset",
        "OGBLDDIDataset",
        "OGBLCitation2Dataset",
        "OGBGMOLHIVDataset",
        "OGBGMOLPCBADataset",
        "OGBGPPADataset",
        "OGBGCode2Dataset",
        "GTNACMDataset",
        "GTNDBLPDataset",
        "GTNIMDBDataset",
        "BlogCatalogDataset",
        "WIKIPEDIADataset"
    ]
else:
    __all__ = [
        "CoraDataset",
        "CiteSeerDataset",
        "PubMedDataset",
        "RedditDataset",
        "AmazonComputersDataset",
        "AmazonPhotoDataset",
        "CoauthorPhysicsDataset",
        "CoauthorCSDataset",
        "MUTAGDataset",
        "ENZYMESDataset",
        "IMDBBinaryDataset",
        "IMDBMultiDataset",
        "RedditBinaryDataset",
        "REDDITMulti5KDataset",
        "COLLABDataset",
        "ProteinsDataset",
        "PTCMRDataset",
        "NCI1Dataset",
        "ACMHANDataset",
        "ACMHGTDataset",
        "OGBNProductsDataset",
        "OGBNProteinsDataset",
        "OGBNArxivDataset",
        "OGBNPapers100MDataset",
        "OGBLPPADataset",
        "OGBLCOLLABDataset",
        "OGBLDDIDataset",
        "OGBLCitation2Dataset",
        "OGBGMOLHIVDataset",
        "OGBGMOLPCBADataset",
        "OGBGPPADataset",
        "OGBGCode2Dataset",
        "GTNACMDataset",
        "GTNDBLPDataset",
        "GTNIMDBDataset",
        "BlogCatalogDataset",
        "WIKIPEDIADataset"
    ]



# Modified from deeprobust Library
import scipy.sparse as sp
import numpy as np
from numba import njit
def truncatedSVD(data, k=50):
    """Truncated SVD on input data.

    Parameters
    ----------
    data :
        input matrix to be decomposed
    k : int
        number of singular values and vectors to compute.

    Returns
    -------
    numpy.array
        reconstructed matrix.
    """
    print('=== GCN-SVD: rank={} ==='.format(k))
    if sp.issparse(data):
        data = data.asfptype()
        U, S, V = sp.linalg.svds(data, k=k)
        print("rank_after = {}".format(len(S.nonzero()[0])))
        diag_S = np.diag(S)
    else:
        U, S, V = np.linalg.svd(data)
        U = U[:, :k]
        S = S[:k]
        V = V[:k, :]
        print("rank_before = {}".format(len(S.nonzero()[0])))
        diag_S = np.diag(S)
        print("rank_after = {}".format(len(diag_S.nonzero()[0])))

    return U @ diag_S @ V

from .._data_preprocessor_registry import DataPreprocessorUniversalRegistry
from ._structure_engineer import StructureEngineer,get_feature,get_edges,set_edges,to_adjacency_list,to_adjacency_matrix,LOGGER
import torch
@DataPreprocessorUniversalRegistry.register_data_preprocessor("gcnsvd")
class GCNSVD(StructureEngineer):
    """GCNSVD uses Truncated SVD as preprocessing.See more details in All You Need Is Low (Rank): Defending
    Against Adversarial Attacks on Graphs,
    https://dl.acm.org/doi/abs/10.1145/3336191.3371789.
    """
    def __init__(self, k=50, threshold=0.05, *args, **kwargs):
        """perform rank-k approximation of adjacency matrix via
        truncated SVD

        Parameters
        ----------
        k : int
            number of singular values and vectors to compute.

        threshold : float
            edges with scores larger than threshold will be kept.
        """
        super(GCNSVD, self).__init__(*args, **kwargs)
        self.k=k
        self.threshold=threshold

    def _transform(self,data):
        adj = get_edges(data) # edge list
        LOGGER.info(f'before modified: {adj.shape}')
        adj = to_adjacency_matrix(adj).numpy() # adjacency matrix
        modified_adj = truncatedSVD(adj,self.k) # adjacency matrix
        modified_adj = (modified_adj> self.threshold).astype(int)
        modified_adj = to_adjacency_list(torch.Tensor(modified_adj)) # edge list
        LOGGER.info(f'after modified: {modified_adj.shape}' )
        set_edges(data,modified_adj)
        return data




import numpy as np
import typing as _typing

import torch
import pickle

from autogl.module.model.encoders.base_encoder import AutoHomogeneousEncoderMaintainer

from ..model import (
    EncoderUniversalRegistry,
    DecoderUniversalRegistry,
    BaseEncoderMaintainer,
    BaseDecoderMaintainer,
    BaseAutoModel,
    ModelUniversalRegistry
)
from ..hpo import AutoModule
import logging
from .evaluation import Evaluation, get_feval, Acc
from ...utils import get_logger

LOGGER_ES = get_logger("early-stopping")

class _DummyModel(torch.nn.Module):
    def __init__(self, encoder: _typing.Union[BaseEncoderMaintainer, BaseAutoModel], decoder: _typing.Optional[BaseDecoderMaintainer]):
        super().__init__()
        if isinstance(encoder, BaseAutoModel):
            self.encoder = encoder.model
            self.decoder = None
        else:
            self.encoder = encoder.encoder
            self.decoder = None if decoder is None else decoder.decoder

    def __str__(self, ):
        return "DummyModel(encoder={}, decoder={})".format(self.encoder, self.decoder)

    def encode(self, *args, **kwargs):
        return self.encoder(*args, **kwargs)
    
    def decode(self, *args, **kwargs):
        if self.decoder is None: return args[0]
        return self.decoder(*args, **kwargs)
    
    def forward(self, *args, **kwargs):
        res = self.encode(*args, **kwargs)
        return self.decode(res, *args, **kwargs)

class EarlyStopping:
    """Early stops the training if validation loss doesn't improve after a given patience."""

    def __init__(
        self,
        patience=7,
        verbose=False,
        delta=0,
        path="checkpoint.pt",
        trace_func=LOGGER_ES.info,
    ):
        """
        Args:
            patience (int): How long to wait after last time validation loss improved.
                            Default: 7
            verbose (bool): If True, prints a message for each validation loss improvement.
                            Default: False
            delta (float): Minimum change in the monitored quantity to qualify as an improvement.
                            Default: 0
            path (str): Path for the checkpoint to be saved to.
                            Default: 'checkpoint.pt'
            trace_func (function): trace print function.
                            Default: print
        """
        self.patience = 100 if patience is None else patience
        self.verbose = verbose
        self.counter = 0
        self.best_score = None
        self.early_stop = False
        self.val_loss_min = np.Inf
        self.delta = delta
        self.path = path
        self.trace_func = trace_func

    def __call__(self, val_loss, model):
        score = -val_loss

        if self.best_score is None:
            self.best_score = score
            self.save_checkpoint(val_loss, model)
        elif score <= self.best_score + self.delta:
            self.counter += 1
            if self.verbose is True:
                self.trace_func(
                    f"EarlyStopping counter: {self.counter} out of {self.patience}"
                )
            if self.counter >= self.patience:
                self.early_stop = True
        else:
            self.best_score = score
            self.save_checkpoint(val_loss, model)
            self.counter = 0

    def save_checkpoint(self, val_loss, model):
        """Saves model when validation loss decrease."""
        if self.verbose:
            self.trace_func(
                f"Validation loss decreased ({self.val_loss_min:.6f} --> {val_loss:.6f}).  Saving model ..."
            )
        self.best_param = pickle.dumps(model.state_dict())
        # torch.save(model.state_dict(), self.path)
        self.val_loss_min = val_loss

    def load_checkpoint(self, model):
        """Load models"""
        if hasattr(self, "best_param"):
            model.load_state_dict(pickle.loads(self.best_param))
        else:
            LOGGER_ES.warn("try to load checkpoint while no checkpoint is saved")


class BaseTrainer(AutoModule):
    def __init__(
        self,
        encoder: _typing.Union[BaseAutoModel, BaseEncoderMaintainer, None],
        decoder: _typing.Union[BaseDecoderMaintainer, None],
        device: _typing.Union[torch.device, str],
        feval: _typing.Union[
            _typing.Sequence[str], _typing.Sequence[_typing.Type[Evaluation]]
        ] = (Acc,),
        loss: str = "nll_loss",
    ):
        """
        The basic trainer.
        Used to automatically train the problems, e.g., node classification, graph classification, etc.
        Parameters
        ----------
        model: `BaseModel` or `str`
            The (name of) model used to train and predict.
        init: `bool`
            If True(False), the model will (not) be initialized.
        """
        super().__init__(device)
        self.encoder = encoder
        self.decoder = None if isinstance(encoder, BaseAutoModel) else decoder
        self.feval = feval
        self.loss = loss

    def _compose_model(self):
        return _DummyModel(self.encoder, self.decoder).to(self.device)

    def _initialize(self):
        self.encoder.initialize()
        if self.decoder is not None:
            self.decoder.initialize(self.encoder)

    @property
    def feval(self) -> _typing.Sequence[_typing.Type[Evaluation]]:
        return self.__feval

    @feval.setter
    def feval(
        self,
        _feval: _typing.Union[
            _typing.Sequence[str], _typing.Sequence[_typing.Type[Evaluation]]
        ],
    ):
        self.__feval: _typing.Sequence[_typing.Type[Evaluation]] = get_feval(_feval)

    @property
    def model(self):
        # compatible with v0.2
        return self.encoder
    
    @model.setter
    def model(self, model):
        # compatible with v0.2
        self.encoder = model

    def to(self, device: _typing.Union[str, torch.device]):
        """
        Transfer the trainer to another device
        Parameters
        ----------
        device: `str` or `torch.device`
            The device this trainer will use
        """
        self.device = device
        if self.encoder is not None:
            self.encoder.to_device(self.device)
        if self.decoder is not None:
            self.decoder.to_device(self.device)

    def get_feval(
        self, return_major: bool = False
    ) -> _typing.Union[
        _typing.Type[Evaluation], _typing.Sequence[_typing.Type[Evaluation]]
    ]:
        """
        Parameters
        ----------
        return_major: ``bool``
            Wether to return the major ``feval``. Default ``False``.
        Returns
        -------
        ``evaluation`` or list of ``evaluation``:
            If ``return_major=True``, will return the major ``evaluation`` method.
            Otherwise, will return the ``evaluation`` element passed when constructing.
        """
        if return_major:
            if isinstance(self.feval, _typing.Sequence):
                return self.feval[0]
            else:
                return self.feval
        return self.feval

    @classmethod
    def save(cls, instance, path):
        with open(path, "wb") as output:
            pickle.dump(instance, output, pickle.HIGHEST_PROTOCOL)

    @classmethod
    def load(cls, path):
        with open(path, "rb") as inputs:
            instance = pickle.load(inputs)
            return instance

    def duplicate_from_hyper_parameter(self, *args, **kwargs) -> "BaseTrainer":
        """Create a new trainer with the given hyper parameter."""
        raise NotImplementedError()

    def train(self, dataset, keep_valid_result):
        """
        Train on the given dataset.
        Parameters
        ----------
        dataset: The dataset used in training.
        keep_valid_result: `bool`
            If True(False), save the validation result after training.
        Returns
        -------
        """
        raise NotImplementedError()

    def predict(self, dataset, mask=None):
        """
        Predict on the given dataset.
        Parameters
        ----------
        dataset: The dataset used in predicting.
        mask: `train`, `val`, or `test`.
            The dataset mask.
        Returns
        -------
        prediction result
        """
        raise NotImplementedError()

    def predict_proba(self, dataset, mask=None, in_log_format=False):
        """
        Predict the probability on the given dataset.
        Parameters
        ----------
        dataset: The dataset used in predicting.
        mask: `train`, `val`, or `test`.
            The dataset mask.
        in_log_format: `bool`.
            If True(False), the probability will (not) be log format.
        Returns
        -------
        The prediction result.
        """
        raise NotImplementedError()

    def get_valid_predict_proba(self):
        """Get the valid result (prediction probability)."""
        raise NotImplementedError()

    def get_valid_predict(self):
        """Get the valid result."""
        raise NotImplementedError()

    def get_valid_score(self, return_major=True):
        """Get the validation score."""
        raise NotImplementedError()

    def __repr__(self) -> str:
        raise NotImplementedError

    def evaluate(self, dataset, mask=None, feval=None):
        """
        Parameters
        ----------
        dataset: The dataset used in evaluation.
        mask: `train`, `val`, or `test`.
            The dataset mask.
        feval: The evaluation methods.
        Returns
        -------
        The evaluation result.
        """
        raise NotImplementedError

    def update_parameters(self, **kwargs):
        """
        Update parameters of this trainer
        """
        for k, v in kwargs.items():
            if k == "feval":
                self.feval = get_feval(v)
            elif k == "device":
                self.to(v)
            elif hasattr(self, k):
                setattr(self, k, v)
            else:
                raise KeyError("Cannot set parameter", k, "for trainer", self.__class__)

    def combined_hyper_parameter_space(self):
        return {
            "trainer": self.hyper_parameter_space,
            "encoder": self.encoder.hyper_parameter_space,
            "decoder": [] if self.decoder is None else self.decoder.hyper_parameter_space
        }


class _BaseClassificationTrainer(BaseTrainer):
    """ Base class of trainer for classification tasks """

    def __init__(
        self,
        encoder: _typing.Union[BaseAutoModel, BaseEncoderMaintainer, str, None],
        decoder: _typing.Union[BaseDecoderMaintainer, str, None],
        num_features: int,
        num_classes: int,
        last_dim: _typing.Union[int, str] = "auto",
        device: _typing.Union[torch.device, str, None] = "auto",
        feval: _typing.Union[
            _typing.Sequence[str], _typing.Sequence[_typing.Type[Evaluation]]
        ] = (Acc,),
        loss: str = "nll_loss",
    ):
        self._encoder = None
        self._decoder = None
        self.num_features = num_features
        self.num_classes = num_classes
        self.last_dim: _typing.Union[int, str] = last_dim
        super(_BaseClassificationTrainer, self).__init__(
            encoder, decoder, device, feval, loss
        )
    
    @property
    def encoder(self):
        return self._encoder
    
    @encoder.setter
    def encoder(self, enc: _typing.Union[BaseAutoModel, BaseEncoderMaintainer, str, None]):
        if isinstance(enc, str):
            if enc in EncoderUniversalRegistry:
                self._encoder = EncoderUniversalRegistry.get_encoder(enc)(
                    self.num_features, final_dimension=self.last_dim, device=self.device, init=self.initialized
                )
            else:
                self._encoder = ModelUniversalRegistry.get_model(enc)(
                    self.num_features, final_dimension=self.last_dim, device=self.device
                )
                
        elif isinstance(enc, BaseEncoderMaintainer):
            self._encoder = enc
        elif isinstance(enc, BaseAutoModel):
            self._encoder = enc
            if self.decoder is not None:
                logging.warn("will disable decoder since a whole model is passed")
                self.decoder = None
        elif enc is None:
            self._encoder = None
        else:
            raise ValueError("Enc {} is not supported!".format(enc))
        self.num_features = self.num_features
        self.num_classes = self.num_classes
        self.last_dim = self.last_dim

    @property
    def decoder(self):
        return self._decoder
    
    @decoder.setter
    def decoder(self, dec: _typing.Union[BaseDecoderMaintainer, str, None]):
        if isinstance(self.encoder, BaseAutoModel):
            logging.warn("Ignore passed dec since enc is a whole model")
            self._decoder = None
            return
        if isinstance(dec, str):
            self._decoder = DecoderUniversalRegistry.get_decoder(dec)(
                self.num_classes, input_dimension=self.last_dim, device=self.device, init=self.initialized
            )
        elif isinstance(dec, BaseDecoderMaintainer):
            self._decoder = dec
        elif dec is None:
            self._decoder = None
        else:
            raise ValueError("Dec {} is not supported!".format(dec))
        self.num_features = self.num_features
        self.num_classes = self.num_classes
        self.last_dim = self.last_dim
    
    @property
    def num_classes(self):
        return self.__num_classes

    @num_classes.setter
    def num_classes(self, num_classes):
        self.__num_classes = num_classes
        if isinstance(self.encoder, BaseAutoModel):
            self.encoder.output_dimension = num_classes
        elif isinstance(self.decoder, BaseDecoderMaintainer):
            self.decoder.output_dimension = num_classes

    @property
    def last_dim(self):
        return self._last_dim
    
    @last_dim.setter
    def last_dim(self, dim):
        self._last_dim = dim
        if isinstance(self.encoder, AutoHomogeneousEncoderMaintainer):
            self.encoder.final_dimension = self._last_dim

    @property
    def num_features(self):
        return self._num_features
    
    @num_features.setter
    def num_features(self, num_features):
        self._num_features = num_features
        if self.encoder is not None:
            self.encoder.input_dimension = num_features

class BaseNodeClassificationTrainer(_BaseClassificationTrainer):
    def __init__(
        self,
        encoder: _typing.Union[BaseAutoModel, BaseEncoderMaintainer, str, None],
        decoder: _typing.Union[BaseDecoderMaintainer, str, None],
        num_features: int,
        num_classes: int,
        device: _typing.Union[torch.device, str, None] = None,
        feval: _typing.Union[
            _typing.Sequence[str], _typing.Sequence[_typing.Type[Evaluation]]
        ] = (Acc,),
        loss: str = "nll_loss",
    ):
        super(BaseNodeClassificationTrainer, self).__init__(
            encoder, decoder, num_features, num_classes, num_classes, device, feval, loss
        )
    
    # override num_classes property to support last_dim setting
    @property
    def num_classes(self):
        return self.__num_classes
    
    @num_classes.setter
    def num_classes(self, num_classes):
        self.__num_classes = num_classes
        if isinstance(self.encoder, BaseAutoModel):
            self.encoder.output_dimension = num_classes
        elif isinstance(self.decoder, BaseDecoderMaintainer):
            self.decoder.output_dimension = num_classes
        self.last_dim = num_classes


class BaseGraphClassificationTrainer(_BaseClassificationTrainer):
    def __init__(
        self,
        encoder: _typing.Union[BaseAutoModel, BaseEncoderMaintainer, str, None] = None,
        decoder: _typing.Union[BaseDecoderMaintainer, str, None] = None,
        num_features: _typing.Optional[int] = None,
        num_classes: _typing.Optional[int] = None,
        num_graph_features: int = 0,
        last_dim: _typing.Union[int, str] = "auto",
        device: _typing.Union[torch.device, str, None] = None,
        feval: _typing.Union[
            _typing.Sequence[str], _typing.Sequence[_typing.Type[Evaluation]]
        ] = (Acc,),
        loss: str = "nll_loss",
    ):
        self._encoder = None
        self._decoder = None
        self.num_graph_features: int = num_graph_features
        super(BaseGraphClassificationTrainer, self).__init__(
            encoder, decoder, num_features, num_classes, last_dim, device, feval, loss
        )

    # override encoder and decoder to depend on graph level features
    @property
    def encoder(self):
        return self._encoder
    
    @encoder.setter
    def encoder(self, enc: _typing.Union[BaseAutoModel, BaseEncoderMaintainer, str, None]):
        if isinstance(enc, str):
            if enc in EncoderUniversalRegistry:
                self._encoder = EncoderUniversalRegistry.get_encoder(enc)(
                    self.num_features,
                    last_dim=self.last_dim,
                    num_graph_features=self.num_graph_features,
                    device=self.device,
                    init=self.initialized
                )
            else:
                self._encoder = ModelUniversalRegistry.get_model(enc)(
                    self.num_features,
                    self.last_dim,
                    device=self.device,
                    num_graph_features=self.num_graph_features,
                )

        elif isinstance(enc, (BaseAutoModel, BaseEncoderMaintainer)):
            self._encoder = enc
            if isinstance(enc, BaseAutoModel) and self.decoder is not None:
                logging.warn("will disable decoder since a whole model is passed")
                self.decoder = None
        elif enc is None:
            self._encoder = None
        else:
            raise ValueError("Enc {} is not supported!".format(enc))
        self.num_features = self.num_features
        self.num_classes = self.num_classes
        self.last_dim = self.last_dim
        self.num_graph_features = self.num_graph_features


    @property
    def decoder(self):
        if isinstance(self.encoder, BaseAutoModel): return None
        return self._decoder
    
    @decoder.setter
    def decoder(self, dec: _typing.Union[BaseDecoderMaintainer, str, None]):
        if isinstance(self.encoder, BaseAutoModel):
            logging.warn("Ignore passed dec since enc is a whole model")
            self._decoder = None
            return
        if isinstance(dec, str):
            self._decoder = DecoderUniversalRegistry.get_decoder(dec)(
                self.num_classes,
                input_dim=self.last_dim,
                num_graph_features=self.num_graph_features,
                device=self.device,
                init=self.initialized
            )
        elif isinstance(dec, (BaseDecoderMaintainer, None)):
            self._decoder = dec
        else:
            raise ValueError("Invalid decoder setting")
        self.num_features = self.num_features
        self.num_classes = self.num_classes
        self.last_dim = self.last_dim

    # override num_classes property to support last_dim setting
    @property
    def num_classes(self):
        return self.__num_classes
    
    @num_classes.setter
    def num_classes(self, num_classes):
        self.__num_classes = num_classes
        if isinstance(self.encoder, BaseAutoModel):
            self.encoder.output_dimension = num_classes
        elif isinstance(self.decoder, BaseDecoderMaintainer):
            self.decoder.output_dimension = num_classes

    @property
    def num_graph_features(self):
        return self._num_graph_features
    
    @num_graph_features.setter
    def num_graph_features(self, num_graph_features):
        self._num_graph_features = num_graph_features
        if self.encoder is not None: self.encoder.num_graph_features = self._num_graph_features
        if self.decoder is not None: self.decoder.num_graph_features = self._num_graph_features


# TODO: according to discussion, link prediction may not belong to classification tasks
class BaseLinkPredictionTrainer(_BaseClassificationTrainer):
    def __init__(
        self,
        encoder: _typing.Union[BaseAutoModel, BaseEncoderMaintainer, str, None] = None,
        decoder: _typing.Union[BaseDecoderMaintainer, str, None] = None,
        num_features: _typing.Optional[int] = None,
        last_dim: _typing.Union[int, str] = "auto",
        device: _typing.Union[torch.device, str, None] = None,
        feval: _typing.Union[
            _typing.Sequence[str], _typing.Sequence[_typing.Type[Evaluation]]
        ] = (Acc,),
        loss: str = "nll_loss",
    ):
        super(BaseLinkPredictionTrainer, self).__init__(
            encoder, decoder, num_features, 2, last_dim, device, feval, loss
        )

    # override decoder since no num_classes is needed
    @property
    def decoder(self):
        if isinstance(self.encoder, BaseAutoModel): return None
        return self._decoder
    
    @decoder.setter
    def decoder(self, dec: _typing.Union[BaseDecoderMaintainer, str, None]):
        if isinstance(self.encoder, BaseAutoModel):
            logging.warn("Ignore passed dec since enc is a whole model")
            self._decoder = None
            return
        if isinstance(dec, str):
            self._decoder = DecoderUniversalRegistry.get_decoder(dec)(
                input_dim=self.last_dim,
                device=self.device,
                init=self.initialized
            )
        elif isinstance(dec, BaseDecoderMaintainer):
            self._decoder = dec
        elif dec is None:
            self._decoder = None
        else:
            raise ValueError("Invalid decoder setting")
        self.num_features = self.num_features
        self.num_classes = self.num_classes
        self.last_dim = self.last_dim


# ============== Het =================
class BaseNodeClassificationHetTrainer(BaseNodeClassificationTrainer):
    """ Base class of trainer for classification tasks """

    def __init__(
        self,
        model: _typing.Union[BaseAutoModel, str],
        dataset: None,
        num_features: int,
        num_classes: int,
        device: _typing.Union[torch.device, str, None] = "auto",
        feval: _typing.Union[
            _typing.Sequence[str], _typing.Sequence[_typing.Type[Evaluation]]
        ] = (Acc,),
        loss: str = "nll_loss",
    ):
        self._dataset = dataset
        super(BaseNodeClassificationHetTrainer, self).__init__(
            model, None, num_features, num_classes, device, feval, loss
        )
        self.from_dataset(dataset)
    
    def from_dataset(self, dataset):
        self._dataset = dataset
        if self.encoder is not None:
            self.encoder.from_dataset(self._dataset)

    @property
    def encoder(self):
        return self._encoder

    @encoder.setter
    def encoder(self, enc: _typing.Union[BaseAutoModel, str, None]):
        if isinstance(enc, str):
            self._encoder = ModelUniversalRegistry.get_model(enc)(
                self.num_features, self.num_classes, device=self.device
            )
        elif isinstance(enc, BaseAutoModel):
            self._encoder = enc
            if self.decoder is not None:
                logging.warn("will disable decoder since a whole model is passed")
                self.decoder = None
        elif enc is None:
            self._encoder = None
        else:
            raise ValueError("Enc {} is not supported!".format(enc))
        self.num_features = self.num_features
        self.num_classes = self.num_classes
        if self._dataset is not None:
            self.from_dataset(self._dataset)



from ._general import StaticGraphToGeneralData, static_graph_to_general_data
from ._nx import (
    HomogeneousStaticGraphToNetworkX
)

try:
    import dgl
except ModuleNotFoundError:
    dgl = None
else:
    from ._dgl import (
        DGLGraphToGeneralStaticGraph, dgl_graph_to_general_static_graph,
        GeneralStaticGraphToDGLGraph, general_static_graph_to_dgl_graph,
    )
try:
    import torch_geometric
except ModuleNotFoundError:
    torch_geometric = None
else:
    from ._pyg import StaticGraphToPyGData, static_graph_to_pyg_data



from .base import BaseEstimator

NAS_ESTIMATOR_DICT = {}


def register_nas_estimator(name):
    def register_nas_estimator_cls(cls):
        if name in NAS_ESTIMATOR_DICT:
            raise ValueError(
                "Cannot register duplicate NAS estimator ({})".format(name)
            )
        if not issubclass(cls, BaseEstimator):
            raise ValueError(
                "Model ({}: {}) must extend NAS estimator".format(name, cls.__name__)
            )
        NAS_ESTIMATOR_DICT[name] = cls
        return cls

    return register_nas_estimator_cls


from .one_shot import OneShotEstimator, OneShotEstimator_HardwareAware 
from .grna_estimator import GRNAEstimator
from .train_scratch import TrainEstimator, TrainEstimator_HardwareAware

def build_nas_estimator_from_name(name: str) -> BaseEstimator:
    """
    Parameters
    ----------
    name: ``str``
        the name of nas estimator.

    Returns
    -------
    BaseEstimator:
        the NAS estimator built using default parameters

    Raises
    ------
    AssertionError
        If an invalid name is passed in
    """
    assert name in NAS_ESTIMATOR_DICT, "HPO module do not have name " + name
    return NAS_ESTIMATOR_DICT[name]()


__all__ = ["BaseEstimator", "OneShotEstimator", "TrainEstimator", "OneShotEstimator_HardwareAware", "TrainEstimator_HardwareAware", "GRNAEstimator"]



import importlib
import os
from .base import BaseSpace

NAS_SPACE_DICT = {}


def register_nas_space(name):
    def register_nas_space_cls(cls):
        if name in NAS_SPACE_DICT:
            raise ValueError("Cannot register duplicate NAS space ({})".format(name))
        if not issubclass(cls, BaseSpace):
            raise ValueError(
                "Model ({}: {}) must extend NAS space".format(name, cls.__name__)
            )
        NAS_SPACE_DICT[name] = cls
        return cls

    return register_nas_space_cls


from .graph_nas_macro import GraphNasMacroNodeClassificationSpace
from .graph_nas import GraphNasNodeClassificationSpace
from .single_path import SinglePathNodeClassificationSpace
from ..backend import *
if not is_dgl():
    from .gasso import GassoSpace
    from .autoattend import AutoAttendNodeClassificationSpace
    from .grna import GRNASpace

def build_nas_space_from_name(name: str) -> BaseSpace:
    """
    Parameters
    ----------
    name: ``str``
        the name of nas space.

    Returns
    -------
    BaseSpace:
        the NAS space built using default parameters

    Raises
    ------
    AssertionError
        If an invalid name is passed in
    """
    assert name in NAS_SPACE_DICT, "HPO module do not have name " + name
    return NAS_SPACE_DICT[name]()


__all__ = [
    "BaseSpace",
    "GraphNasMacroNodeClassificationSpace",
    "GraphNasNodeClassificationSpace",
    "SinglePathNodeClassificationSpace",
]

if not is_dgl():
    __all__.append("GassoSpace")
    __all__.append("GRNASpace")
    __all__.append("AutoAttendNodeClassificationSpace")




"""
Auto classifier for classification problems.
"""

from .base import BaseClassifier
from .graph_classifier import AutoGraphClassifier
from .node_classifier import AutoNodeClassifier
from .link_predictor import AutoLinkPredictor
from autogl.backend import DependentBackend
if DependentBackend.is_dgl():
    from .hetero import AutoHeteroNodeClassifier
if DependentBackend.is_pyg():
    from .ssl import SSLGraphClassifier

__all__ = [
    "BaseClassifier",
    "AutoGraphClassifier",
    "AutoNodeClassifier",
    "AutoLinkPredictor",
]

if DependentBackend.is_dgl():
    __all__.extend(['AutoHeteroNodeClassifier'])

if DependentBackend.is_pyg():
    __all__.extend(['SSLGraphClassifier'])
